import pytest
import numpy as np
import matplotlib.pyplot as plt  # this import is needed for interactive tests

from locan import RenderEngine  # this import is needed for interactive tests
from locan.dependencies import HAS_DEPENDENCY
from locan import render_2d_mpl, render_2d_napari, \
    select_by_drawing_napari, render_2d_rgb_napari
from locan import render_2d
from locan import transform_affine

if HAS_DEPENDENCY["napari"]: import napari


pytestmark = pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="requires napari")

HAS_NAPARI_AND_PYTESTQT = HAS_DEPENDENCY["napari"] and HAS_DEPENDENCY["pytestqt"]
# pytestqt is not a requested or extra dependency.
# If napari and pytest-qt is installed, all tests run.
# Tests in docker or GitHub actions on linux require xvfb for tests with pytest-qt to run.


@pytest.mark.gui
@pytest.mark.parametrize("test_input, expected", list((member, 0) for member in list(RenderEngine)))
def test_render_2d_gui(locdata_blobs_2d, test_input, expected):
    if test_input == RenderEngine.NAPARI:
        render_2d(locdata_blobs_2d, render_engine=test_input)
        # napari.run()
    else:
        render_2d(locdata_blobs_2d, render_engine=test_input)
    # plt.show()

    plt.close('all')


@pytest.mark.visual
# this is to check overlay of rendered image and single localization points
def test_render_2d_napari_coordinates(locdata_blobs_2d):
    render_2d_mpl(locdata_blobs_2d, bin_size=10, cmap='viridis')
    plt.show()

    viewer, bins = render_2d_napari(locdata_blobs_2d, bin_size=10, cmap='viridis', gamma=0.1)
    viewer.add_points((locdata_blobs_2d.coordinates - np.array(bins.bin_range)[:, 0]) / 10 )
    napari.run()

    plt.close('all')


@pytest.mark.gui
def test_render_2d_napari_gui(locdata_blobs_2d):
    render_2d_mpl(locdata_blobs_2d, bin_size=100, cmap='viridis')
    plt.show()

    render_2d_napari(locdata_blobs_2d, bin_size=100, cmap='viridis', gamma=0.1)
    napari.run()

    viewer, _ = render_2d_napari(locdata_blobs_2d, bin_size=50, cmap='magenta', gamma=0.1)
    render_2d_napari(locdata_blobs_2d, viewer=viewer, bin_size=100, cmap='cyan', gamma=0.1, scale=(2, 2),
                     blending='additive')
    # napari.run()

    render_2d(locdata_blobs_2d, render_engine=RenderEngine.NAPARI)
    # napari.run()

    plt.close('all')


@pytest.mark.skipif(not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt.")
def test_render_2d_napari(make_napari_viewer, locdata_blobs_2d):
    viewer = make_napari_viewer()
    render_2d_napari(locdata_blobs_2d, viewer=viewer, bin_size=100, cmap='viridis', gamma=0.1)
    assert len(viewer.layers) == 1
    assert viewer.layers[0].name == 'LocData 0'
    viewer.close()


@pytest.mark.gui
def test_render_2d_napari_empty_gui(locdata_empty):
    render_2d_napari(locdata_empty, bin_size=100, cmap='viridis', gamma=0.1)
    napari.run()


@pytest.mark.skipif(not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt.")
def test_render_2d_napari_empty(make_napari_viewer, locdata_empty):
    viewer = make_napari_viewer()
    render_2d_napari(locdata_empty, viewer=viewer, bin_size=100, cmap='viridis', gamma=0.1)
    assert viewer.layers == []
    viewer.close()


@pytest.mark.gui
def test_render_2d_napari_single_gui(locdata_single_localization, caplog):
    render_2d_napari(locdata_single_localization, bin_size=100, cmap='viridis', gamma=0.1)
    assert caplog.record_tuples[1] == ('locan.render.render2d', 30, 'Locdata carries a single localization.')
    napari.run()


@pytest.mark.skipif(not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt.")
def test_render_2d_napari_single(make_napari_viewer, locdata_single_localization, caplog):
    viewer = make_napari_viewer()
    render_2d_napari(locdata_single_localization, viewer=viewer,
                     bin_size=100, cmap='viridis', gamma=0.1)
    viewer.close()
    assert caplog.record_tuples[0] == ('locan.render.render2d', 30, 'Locdata carries a single localization.')


@pytest.mark.gui
def test_select_by_drawing_napari_gui(locdata_blobs_2d):
    viewer = napari.Viewer()
    viewer.add_shapes(data=np.array([(1, 10), (10, 20)]), shape_type='rectangle')

    rois = select_by_drawing_napari(locdata_blobs_2d, viewer=viewer, bin_size=10, cmap='viridis', gamma=0.1)
    # No need for napari.run() since it is called inside select_by_drawing_napari.
    print(rois)
    print(viewer.layers['Shapes'].data)
    assert len(rois) == 1
    assert repr(rois[0].region) == 'Rectangle((122.0, 565.0), 90.0, 100.0, 0)'


@pytest.mark.skipif(not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt.")
def test_select_by_drawing_napari(make_napari_viewer, locdata_blobs_2d):
    viewer = make_napari_viewer()
    viewer.add_shapes(data=np.array([(1, 10), (10, 20)]), shape_type='rectangle')

    rois = select_by_drawing_napari(locdata_blobs_2d, viewer=viewer, napari_run=False,
                                    bin_size=10, cmap='viridis', gamma=0.1)
    assert len(rois) == 1
    assert repr(rois[0].region) == 'Rectangle((122.0, 565.0), 90.0, 100.0, 0)'
    viewer.close()


@pytest.mark.gui
def test_select_by_drawing_napari_2(locdata_blobs_2d):
    roi_list = select_by_drawing_napari(locdata_blobs_2d)
    print(roi_list)


@pytest.mark.gui
def test_render_2d_rgb_napari_gui(locdata_blobs_2d):
    locdata_0 = locdata_blobs_2d
    locdata_1 = transform_affine(locdata_blobs_2d, offset=(20, 0))
    render_2d_rgb_napari([locdata_0, locdata_1], bin_size=20)

    napari.run()


@pytest.mark.skipif(not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt.")
def test_render_2d_rgb_napari(make_napari_viewer, locdata_blobs_2d):
    viewer = make_napari_viewer()
    locdata_0 = locdata_blobs_2d
    locdata_1 = transform_affine(locdata_blobs_2d, offset=(20, 0))
    render_2d_rgb_napari([locdata_0, locdata_1], viewer=viewer, bin_size=20)
    viewer.close()



from pathlib import Path
import tempfile
import pickle
from io import StringIO
import logging

import pytest
from pandas.testing import assert_frame_equal

import locan.constants
from locan.dependencies import HAS_DEPENDENCY
from locan.data import metadata_pb2
from locan.locan_io import load_txt_file, load_locdata

from locan.locan_io.locdata.io_locdata import _map_file_type_to_load_function


def test_loading_txt_file(caplog):
    dat = load_txt_file(path=locan.ROOT_DIR / 'tests/test_data/five_blobs.txt', nrows=10)
    # print(dat.data)
    assert (len(dat) == 10)

    dat = load_txt_file(path=locan.ROOT_DIR / 'tests/test_data/five_blobs.txt',
                           columns=['index', 'position_x', 'position_y', 'cluster_label', 'frame'], nrows=10)
    # print(dat.data)
    assert (len(dat) == 10)

    file_like = StringIO("index,position_x,position_y,cluster_label\n"
                         "0,624,919,3")
    dat = load_txt_file(path=file_like,
                           columns=['index', 'position_x', 'position_y', 'cluster_label'], nrows=1)
    assert (len(dat) == 1)

    dat = load_txt_file(path=locan.ROOT_DIR / 'tests/test_data/five_blobs.txt',
                           columns=['c1'], nrows=10)
    # print(dat.data)
    assert caplog.record_tuples == [('locan.locan_io.locdata.utilities', logging.WARNING,
                                     'Column c1 is not a Locan property standard.')]

    assert (len(dat) == 10)

    dat = load_txt_file(path=locan.ROOT_DIR / 'tests/test_data/five_blobs.txt',
                           columns=['c1'], nrows=10, property_mapping={'c1': 'something'})
    assert list(dat.data.columns) == ['something']


def test__map_file_type_to_load_function():
    file_type = _map_file_type_to_load_function(file_type=1)
    assert callable(file_type)
    file_type = _map_file_type_to_load_function(file_type='RAPIDSTORM')
    assert callable(file_type)
    file_type = _map_file_type_to_load_function(file_type=locan.constants.FileType.RAPIDSTORM)
    assert callable(file_type)
    file_type = _map_file_type_to_load_function(file_type=metadata_pb2.RAPIDSTORM)
    assert callable(file_type)


def test_load_locdata():
    dat = load_locdata(path=locan.ROOT_DIR / 'tests/test_data/rapidSTORM_dstorm_data.txt',
                       file_type='RAPIDSTORM',
                       nrows=10)
    assert (len(dat) == 10)

    dat = load_locdata(path=locan.ROOT_DIR / 'tests/test_data/Elyra_dstorm_data.txt',
                       file_type='ELYRA',
                       nrows=10)
    assert (len(dat) == 10)

    dat = load_locdata(path=locan.ROOT_DIR / 'tests/test_data/Elyra_dstorm_data.txt',
                       file_type='elyra',
                       nrows=10)
    assert (len(dat) == 10)

    dat = load_locdata(path=locan.ROOT_DIR / 'tests/test_data/Elyra_dstorm_data.txt',
                       file_type=3,
                       nrows=10)
    assert (len(dat) == 10)

    dat = load_locdata(path=locan.ROOT_DIR / 'tests/test_data/Nanoimager_dstorm_data.csv',
                       file_type='NANOIMAGER',
                       nrows=10)
    assert (len(dat) == 10)

    dat = load_locdata(path=locan.ROOT_DIR / 'tests/test_data/rapidSTORM_dstorm_track_data.txt',
                       file_type='RAPIDSTORMTRACK',
                       nrows=10)
    assert (len(dat) == 10)

    dat = load_locdata(path=locan.ROOT_DIR / 'tests/test_data/SMLM_dstorm_data.smlm',
                       file_type='SMLM',
                       nrows=10)
    assert (len(dat) == 10)


@pytest.mark.skipif(not HAS_DEPENDENCY["h5py"], reason="requires h5py.")
def test_load_locdata_2():
    dat = load_locdata(path=locan.ROOT_DIR / 'tests/test_data/decode_dstorm_data.h5',
                       file_type='DECODE',
                       nrows=10)
    assert (len(dat) == 10)

    dat = load_locdata(path=locan.ROOT_DIR / 'tests/test_data/smap_dstorm_data.mat',
                       file_type='SMAP',
                       nrows=10)
    assert (len(dat) == 10)


def test_pickling_locdata(locdata_2d):
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / 'pickled_locdata.pickle'
        with open(file_path, 'wb') as file:
            pickle.dump(locdata_2d, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, 'rb') as file:
            locdata = pickle.load(file)
        assert len(locdata_2d) == len(locdata)
        assert isinstance(locdata.meta, metadata_pb2.Metadata)



"""
Hull objects of localization data.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   hull
   alpha_shape

"""

from locan.data.hulls.hull import *
from locan.data.hulls.alpha_shape import *


__all__ = []
__all__.extend(hull.__all__)
__all__.extend(alpha_shape.__all__)



from pathlib import Path
import tempfile

import pytest
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from locan import LocData, Rectangle, Ellipse, Polygon
from locan import ROOT_DIR
from locan import Roi, rasterize
from locan.data import metadata_pb2


def test_Roi_2d(locdata_2d):
    roi = Roi(reference=locdata_2d, region=Rectangle((0, 0), 2, 2, 0))
    assert repr(roi) == f"Roi(reference={locdata_2d}, region=Rectangle((0, 0), 2, 2, 0),  loc_properties=())"
    assert isinstance(roi._region, Rectangle)
    assert isinstance(roi.region, Rectangle)

    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is roi.region

    roi.region = Ellipse((0, 0), 2, 2, 0)
    assert isinstance(roi.region, Ellipse)

    roi = Roi(reference=locdata_2d, region=Rectangle((0, 0), 4, 4, 0),
              loc_properties=('position_y', 'frame'))
    new_dat = roi.locdata()
    assert len(new_dat) == 2
    assert new_dat.region is None

    locdata_empty = LocData()
    roi_empty = Roi(reference=locdata_empty, region=Rectangle((0, 0), 4, 4, 0))
    empty_dat = roi_empty.locdata()
    assert len(empty_dat) == 0
    assert new_dat.region is None


def test_Roi_3d(locdata_3d):
    roi = Roi(reference=locdata_3d, region=Rectangle((0, 0), 2, 2, 0))
    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is None
    assert isinstance(roi.region, Rectangle)
    roi.region = Ellipse((0, 0), 2, 2, 0)
    assert isinstance(roi.region, Ellipse)

    roi = Roi(reference=locdata_3d, region=Rectangle((0, 0), 2, 3, 0),
              loc_properties=('position_y', 'position_z'))
    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is None


def test_pickling_locdata_from_Roi(locdata_2d):
    import pickle
    roi = Roi(reference=locdata_2d, region=Rectangle((0, 0), 2, 1, 0))
    new_dat = roi.locdata()

    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / 'pickled_locdata.pickle'
        with open(file_path, 'wb') as file:
            pickle.dump(new_dat, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, 'rb') as file:
            locdata = pickle.load(file)
        assert len(new_dat) == len(locdata)
        assert isinstance(locdata.meta, metadata_pb2.Metadata)


def test_Roi_io(locdata_2d):
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / 'roi.yaml'
        #file_path = ROOT_DIR / 'tests/test_data/roi.yaml'

        roi = Roi(reference=locdata_2d, region=Rectangle((0, 0), 2, 1, 10))
        with pytest.warns(UserWarning):
            roi.to_yaml(path=file_path)

        roi = Roi(region=Rectangle((0, 0), 2, 1, 0))
        roi.to_yaml(path=file_path)

        roi_new = Roi.from_yaml(path=file_path)
        assert roi_new.reference is None
        assert isinstance(roi_new.region, Rectangle)
        assert roi_new.loc_properties == []

        roi = Roi(reference=dict(file_path=ROOT_DIR / 'tests/test_data/five_blobs.txt', file_type=1),
                  region=Rectangle((0, 0), 2, 1, 0))
        assert isinstance(roi.reference, (metadata_pb2.Metadata, Path))
        roi.to_yaml(path=file_path)

        roi_new = Roi.from_yaml(path=file_path)
        assert roi_new

        # test region specs with numpy floats
        roi = Roi(reference=dict(file_path=file_path, file_type=1),
                  region=Rectangle((0, 0), float(2), float(1), float(10)))
        assert isinstance(roi.reference, metadata_pb2.Metadata)
        roi.to_yaml(path=file_path)

        roi_new = Roi.from_yaml(path=file_path)
        assert roi_new

        meta_ = metadata_pb2.Metadata()
        meta_.file.path = str(file_path)
        meta_.file.type = 1
        locdata_2 = LocData.from_selection(locdata_2d, meta=meta_)
        roi = Roi(reference=locdata_2, region=Rectangle((0, 0), 2, 1, 0))
        assert isinstance(roi.reference.meta, metadata_pb2.Metadata)
        roi.to_yaml(path=file_path)

        roi_new = Roi.from_yaml(path=file_path)
        assert roi_new
        assert isinstance(roi_new.reference, metadata_pb2.Metadata)


def test_roi_locdata_from_file(locdata_blobs_3d):
    roi = Roi(reference=locdata_blobs_3d, region=Rectangle((100, 500), 200, 200, 10))
    dat = roi.locdata()
    assert(len(dat) == 10)

    roi = Roi(reference=locdata_blobs_3d, region=Rectangle((100, 3), 200, 2, 0),
              loc_properties=['position_x', 'cluster_label'])
    dat = roi.locdata()
    assert(len(dat) == 10)


def test_as_artist():
    roi_rectangle = Roi(reference=None, region=Rectangle((0, 0), 0.5, 0.5, 10))
    roi_ellipse = Roi(reference=None, region=Ellipse((0.5, 0.5), 0.2, 0.3, 10))
    roi_polygon = Roi(reference=None, region=Polygon(((0, 0.8), (0.1, 0.9), (0.6, 0.9), (0.7, 0.7), (0, 0.8))))
    fig, ax = plt.subplots()
    ax.add_patch(roi_rectangle._region.as_artist())
    ax.add_patch(roi_ellipse._region.as_artist())
    ax.add_patch(roi_polygon._region.as_artist())
    # plt.show()

    plt.close('all')


# to be deprecated
# @pytest.mark.gui
# def test_select_by_drawing_mpl():
#     dat = load_txt_file(path=ROOT_DIR / 'tests/test_data/five_blobs.txt')
#     select_by_drawing_mpl(dat, region_type='rectangle')
#     select_by_drawing_mpl(dat, region_type='ellipse')


# standard LocData fixtures
def test_rasterize(locdata_2d, locdata_single_localization, locdata_non_standard_index):
    res = rasterize(locdata=locdata_2d, support=None, n_regions=(5, 2))
    assert isinstance(res[0].region, Rectangle)
    assert repr(res[0].region) == 'Rectangle((1.0, 1.0), 0.8, 2.5, 0)'
    assert len(res) == 10

    res = rasterize(locdata=locdata_2d, support=((0, 10), (10, 20)), n_regions=(5, 2))
    assert isinstance(res[0].region, Rectangle)
    assert repr(res[0].region) == 'Rectangle((0.0, 10.0), 2.0, 5.0, 0)'
    assert len(res) == 10

    res = rasterize(locdata=locdata_single_localization, support=((0, 10), (10, 20)), n_regions=(5, 2))
    assert isinstance(res[0].region, Rectangle)
    assert repr(res[0].region) == 'Rectangle((0.0, 10.0), 2.0, 5.0, 0)'
    assert len(res) == 10

    res = rasterize(locdata=locdata_non_standard_index, support=((0, 10), (10, 20)), n_regions=(5, 2))
    assert isinstance(res[0].region, Rectangle)
    assert repr(res[0].region) == 'Rectangle((0.0, 10.0), 2.0, 5.0, 0)'
    assert len(res) == 10


def test_rasterize_(locdata_empty):
    with pytest.raises(ValueError):
        rasterize(locdata=locdata_empty, support=None, n_regions=(5, 2))
    with pytest.raises(ValueError):
        rasterize(locdata=locdata_empty, support=((0, 10), (10, 20)), n_regions=(5, 2))


def test_rasterize_3d(locdata_3d):
    res = rasterize(locdata=locdata_3d, support=None, n_regions=(5, 2), loc_properties=['position_x', 'position_z'])
    assert isinstance(res[0].region, Rectangle)
    assert repr(res[0].region) == 'Rectangle((1.0, 1.0), 0.8, 2.0, 0)'
    assert res[0].loc_properties == ['position_x', 'position_z']
    assert len(res) == 10



"""
command-line interface
"""
import sys
import argparse

from locan.scripts.script_draw_roi import _add_arguments as _add_arguments_draw_roi
from locan.scripts.script_check import _add_arguments as _add_arguments_check
from locan.scripts.script_rois import _add_arguments as _add_arguments_rois
from locan.scripts.script_napari import _add_arguments as _add_arguments_napari
from locan.scripts.script_show_versions import _add_arguments as _add_arguments_show_versions
from locan.scripts.script_test import _add_arguments as _add_arguments_test


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(description='Entry point for locan.')

    subparsers = parser.add_subparsers(dest='command')

    # parser for the command draw_roi_mpl
    parser_draw_roi = subparsers.add_parser(name='draw_roi_mpl',
                                            description='Set roi by drawing a boundary in mpl.')
    _add_arguments_draw_roi(parser_draw_roi)

    # parser for the command rois
    parser_rois = subparsers.add_parser(name='rois',
                                        description='Define rois by adding shapes in napari.')
    _add_arguments_rois(parser_rois)

    # parser for the command check
    parser_check = subparsers.add_parser(name='check',
                                         description='Show localizations in original recording.')
    _add_arguments_check(parser_check)

    # parser for the command napari
    parser_napari = subparsers.add_parser(name='napari',
                                          description='Render localization data in napari.')
    _add_arguments_napari(parser_napari)

    # parser for the command show_versions
    parser_show_versions = subparsers.add_parser(name='show_versions',
                                                 description='Show system information and dependency versions.')
    _add_arguments_show_versions(parser_show_versions)

    # parser for the command test
    parser_test = subparsers.add_parser(name='test',
                                        description='Run test suite.')
    _add_arguments_test(parser_test)

    # Parse
    returned_args = parser.parse_args(args)

    if returned_args.command:

        if returned_args.command == "draw_roi_mpl":
            from .scripts.script_draw_roi import sc_draw_roi_mpl
            sc_draw_roi_mpl(returned_args.file, returned_args.type, returned_args.roi_file_indicator,
                            returned_args.region_type)

        elif returned_args.command == "rois":
            from .scripts.script_rois import sc_draw_roi_napari
            sc_draw_roi_napari(file_path=returned_args.file, file_type=returned_args.type,
                               roi_file_indicator=returned_args.roi_file_indicator, bin_size=returned_args.bin_size,
                               rescale=returned_args.rescale)

        elif returned_args.command == "check":
            from .scripts.script_check import sc_check
            sc_check(pixel_size=returned_args.pixel_size, file_images=returned_args.file_images,
                     file_locdata=returned_args.file_locdata, file_type=returned_args.file_type,
                     transpose=True, kwargs_image={}, kwargs_points={})

        elif returned_args.command == "napari":
            from .scripts.script_napari import sc_napari
            sc_napari(file_path=returned_args.file, file_type=returned_args.type, bin_size=returned_args.bin_size,
                      rescale=returned_args.rescale)

        elif returned_args.command == "show_versions":
            from .scripts.script_show_versions import sc_show_versions
            sc_show_versions(verbose=returned_args.verbose, extra_dependencies=returned_args.extra,
                             other_dependencies=returned_args.other)

        elif returned_args.command == "test":
            from .scripts.script_test import sc_test
            sc_test(args=returned_args.pytest_args)

    else:
        print("This is the command line entry point for locan. Get more information with 'locan -h'.")


if __name__ == "__main__":
    sys.exit(main())



"""

Render localization data.

This module provides functions to render and present localization data.
It mostly makes use of matplotlib and napari.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   render2d
   render3d
   utilities
   transform

"""

from .render2d import *
from .render3d import *
from .utilities import *
from .transform import *

__all__ = []
__all__.extend(render2d.__all__)
__all__.extend(render3d.__all__)
__all__.extend(utilities.__all__)
__all__.extend(transform.__all__)



from copy import deepcopy

import pytest
import numpy as np
import matplotlib.pyplot as plt

from locan import bunwarp, render_2d_mpl, render_2d_rgb_mpl
import locan.constants
from locan.dependencies import HAS_DEPENDENCY
from locan.data.region import Polygon
from locan.locan_io.locdata.io_locdata import load_rapidSTORM_file, load_asdf_file
from locan.data.transform import randomize, transform_affine
from locan.data.transform.transformation import _homogeneous_matrix
from locan.data.transform.bunwarpj import _read_matrix, _unwarp


def test_randomize_2d(locdata_2d):
    locdata_2d = deepcopy(locdata_2d)

    locdata_randomized = randomize(locdata_2d, hull_region='bb')
    # locdata_randomized.print_meta()
    assert len(locdata_randomized) == 6
    # print(locdata_randomized.meta)
    assert locdata_randomized.meta.history[-1].name == 'randomize'

    locdata_randomized = randomize(locdata_2d, hull_region='ch')
    assert len(locdata_randomized) == 6

    locdata_randomized = randomize(locdata_2d, hull_region='obb')
    assert len(locdata_randomized) == 6

    locdata_2d.update_alpha_shape(alpha=10)
    locdata_randomized = randomize(locdata_2d, hull_region='as')
    assert len(locdata_randomized) == 6

    region = Polygon(((0, 0), (0, 5), (4, 3), (2, 0.5), (0, 0)))
    locdata_randomized = randomize(locdata_2d, hull_region=region)
    assert len(locdata_randomized) == 6


def test_randomize_3d(locdata_3d):
    locdata_randomized = randomize(locdata_3d, hull_region='bb')
    assert len(locdata_randomized) == 6
    assert locdata_randomized.meta.history[-1].name == 'randomize'

    # todo: implement make_csr in 3d
    with pytest.raises(NotImplementedError):
        locdata_randomized = randomize(locdata_3d, hull_region='ch')
        assert len(locdata_randomized) == 6

    # region_dict = dict(region='polygon', region_specs=((0, 0, 0), (0, 5, 0), (4, 3, 2), (2, 0.5, 2), (0, 0, 0)))
    # locdata_randomized = randomize(locdata_3d, hull_region=region_dict)
    # assert len(locdata_randomized) == 6


@pytest.mark.parametrize('fixture_name, expected', [
    # ('locdata_empty', 0),
    # ('locdata_single_localization', 1),
    ('locdata_2d', 6),
    ('locdata_3d', 6),
    ('locdata_non_standard_index', 6)
])
def test_randomize_locdata_objects(
        locdata_empty, locdata_single_localization, locdata_2d, locdata_3d, locdata_non_standard_index,
        fixture_name, expected):
    locdata = eval(fixture_name)
    locdata_randomized = randomize(locdata, hull_region='bb')
    assert len(locdata_randomized) == expected


def test_bunwarp_raw_transformation():
    matrix_path = locan.ROOT_DIR / 'tests/test_data/transform/BunwarpJ_transformation_raw_green.txt'
    dat_green = load_asdf_file(path=locan.ROOT_DIR /
                                     'tests/test_data/transform/rapidSTORM_beads_green.asdf')

    matrix_x, matrix_y = _read_matrix(path=matrix_path)
    assert np.array_equal(matrix_x.shape, [140, 140])

    dat_green_flipped = transform_affine(dat_green,
                               matrix=[[-1, 0], [0, 1]],
                               offset=[1400, 0]
                               )
    new_points = _unwarp(dat_green_flipped.coordinates, matrix_x, matrix_y, pixel_size=(10, 10))
    assert len(new_points) == len(dat_green)

    dat_green_transformed = bunwarp(locdata=dat_green, matrix_path=matrix_path, pixel_size=(10, 10), flip=True)
    assert len(dat_green_transformed) == len(dat_green)
    assert np.array_equal(dat_green_transformed.coordinates, new_points)
    assert dat_green_transformed.meta.history[-1].name == 'bunwarp'

    # for visual inspection
    dat_red = load_asdf_file(path=locan.ROOT_DIR /
                                        'tests/test_data/transform/rapidSTORM_beads_red.asdf')

    fig, ax = plt.subplots(1, 1, figsize=(16, 8))
    render_2d_mpl(dat_red, ax=ax, bin_size=5, bin_range=((0, 1000), (0, 1400)), rescale=True, cmap='Reds')
    render_2d_mpl(dat_green, ax=ax, bin_size=5, bin_range=((0, 1000), (0, 1400)), rescale=True, cmap='Greens')
    render_2d_mpl(dat_green_transformed, ax=ax, bin_size=5, bin_range=((0, 1000), (0, 1400)),
                  rescale=True, cmap='Blues', alpha=0.5)
    #  plt.show()

    render_2d_rgb_mpl([dat_red, dat_green_transformed], bin_size=5, bin_range=((0, 1000), (0, 1400)))
    # plt.show()

    plt.close('all')


def test_homogeneous_matrix():
    matrix_out = _homogeneous_matrix()
    result = np.identity(4)
    assert np.array_equal(matrix_out, result)

    matrix = ((1, 2, 3), (4, 5, 6), (7, 8, 9))
    offset = (10, 20, 30)
    matrix_out = _homogeneous_matrix(matrix, offset)
    result = np.array(((1, 2, 3, 10), (4, 5, 6, 20), (7, 8, 9, 30), (0, 0, 0, 1)))
    assert np.array_equal(matrix_out, result)

    matrix = ((1, 2), (3, 4))
    offset = (10, 20)
    matrix_out = _homogeneous_matrix(matrix, offset)
    result = np.array(((1, 2, 10), (3, 4, 20), (0, 0, 1)))
    assert np.array_equal(matrix_out, result)


@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', 0),
    ('locdata_single_localization', 4),
    ('locdata_2d', 4),
    ('locdata_3d', 5),
    ('locdata_non_standard_index', 4)
])
def test_standard_locdata_objects(
        locdata_empty, locdata_single_localization, locdata_2d, locdata_3d, locdata_non_standard_index,
        fixture_name, expected):
    locdata = eval(fixture_name)
    new_locdata = transform_affine(locdata)
    assert len(new_locdata.data.columns) == expected


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
@pytest.mark.parametrize('fixture_name, expected', [
    ('locdata_empty', 0),
    ('locdata_single_localization', 4),
    ('locdata_2d', 4),
    ('locdata_3d', 5),
    ('locdata_non_standard_index', 4)
])
def test_standard_locdata_objects_open3d(
        locdata_empty, locdata_single_localization, locdata_2d, locdata_3d, locdata_non_standard_index,
        fixture_name, expected):
    locdata = eval(fixture_name)
    new_locdata = transform_affine(locdata, method='open3d')
    assert len(new_locdata.data.columns) == expected


def test_transformation_affine_2d(locdata_2d):
    new_locdata = transform_affine(locdata_2d)
    assert np.array_equal(new_locdata.coordinates, locdata_2d.coordinates)
    assert len(new_locdata.data.columns) == 4
    assert new_locdata.meta.history[-1].name == 'transform_affine'
    assert "'matrix': None, 'offset': None, 'pre_translation': None, 'method': 'numpy'" \
           in new_locdata.meta.history[-1].parameter

    matrix = ((-1, 0), (0, -1))
    offset = (10, 10)
    pre_translation = (100, 100)

    new_locdata = transform_affine(locdata_2d, matrix, offset)
    points_target = ((9, 9), (9, 5), (8, 7), (7, 4), (6, 8), (5, 5))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 4

    new_locdata = transform_affine(locdata_2d, offset=offset, pre_translation=pre_translation)
    points_target = ((11, 11), (11, 15), (12, 13), (13, 16), (14, 12), (15, 15))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 4


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test_transformation_affine_2d_open3d(locdata_2d):
    new_locdata = transform_affine(locdata_2d, method='open3d')
    assert np.array_equal(new_locdata.coordinates, locdata_2d.coordinates)
    assert len(new_locdata.data.columns) == 4

    matrix = ((-1, 0), (0, -1))
    offset = (10, 10)
    pre_translation = (100, 100)

    new_locdata = transform_affine(locdata_2d, matrix, offset,  method='open3d')
    points_target = ((9, 9), (9, 5), (8, 7), (7, 4), (6, 8), (5, 5))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 4

    new_locdata = transform_affine(locdata_2d,  offset=offset, pre_translation=pre_translation,  method='open3d')
    points_target = ((11, 11), (11, 15), (12, 13), (13, 16), (14, 12), (15, 15))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 4


def test_transformation_affine_3d(locdata_3d):
    new_locdata = transform_affine(locdata_3d)
    assert np.array_equal(new_locdata.coordinates, locdata_3d.coordinates)
    assert len(new_locdata.data.columns) == 5
    assert new_locdata.meta.history[-1].name == 'transform_affine'
    assert "'matrix': None, 'offset': None, 'pre_translation': None, 'method': 'numpy'" \
           in new_locdata.meta.history[-1].parameter

    matrix = ((-1, 0, 0), (0, -1, 0), (0, 0, -1))
    offset = (10, 10, 10)
    pre_translation = (100, 100, 100)

    new_locdata = transform_affine(locdata_3d, matrix, offset)
    points_target = ((9, 9, 9), (9, 5, 8), (8, 7, 5), (7, 4, 6), (6, 8, 7), (5, 5, 8))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 5

    new_locdata = transform_affine(locdata_3d, offset=offset, pre_translation=pre_translation)
    points_target = ((11, 11, 11), (11, 15, 12), (12, 13, 15), (13, 16, 14), (14, 12, 13), (15, 15, 12))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 5


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test_transformation_affine_3d_open3d(locdata_3d):
    new_locdata = transform_affine(locdata_3d, method='open3d')
    assert np.array_equal(new_locdata.coordinates, locdata_3d.coordinates)
    assert len(new_locdata.data.columns) == 5

    matrix = ((-1, 0, 0), (0, -1, 0), (0, 0, -1))
    offset = (10, 10, 10)
    pre_translation = (100, 100, 100)

    new_locdata = transform_affine(locdata_3d, matrix, offset,  method='open3d')
    points_target = ((9, 9, 9), (9, 5, 8), (8, 7, 5), (7, 4, 6), (6, 8, 7), (5, 5, 8))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 5

    new_locdata = transform_affine(locdata_3d,  offset=offset, pre_translation=pre_translation,  method='open3d')
    points_target = ((11, 11, 11), (11, 15, 12), (12, 13, 15), (13, 16, 14), (14, 12, 13), (15, 15, 12))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 5



import pytest
import numpy as np
import matplotlib.pyplot as plt  # this import is needed for interactive tests

from locan import RenderEngine  # this import is needed for interactive tests
from locan.dependencies import HAS_DEPENDENCY
from locan import render_3d, render_3d_napari, scatter_3d_mpl, render_3d_rgb_napari
from locan import transform_affine

if HAS_DEPENDENCY["napari"]: import napari


@pytest.mark.visual
# this is to check overlay of rendered image and single localization points
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_3d_napari_coordinates(locdata_blobs_3d):
    viewer, histogram = render_3d_napari(locdata_blobs_3d, bin_size=10, cmap='viridis', gamma=0.1)
    viewer.add_points((locdata_blobs_3d.coordinates - np.array(histogram.bins.bin_range)[:, 0]) / 10 )
    napari.run()


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_3d_napari(locdata_blobs_3d):
    render_3d_napari(locdata_blobs_3d, bin_size=100, cmap='viridis', gamma=0.1)
    # napari.run()

    viewer = render_3d_napari(locdata_blobs_3d, bin_size=50, cmap='magenta', gamma=0.1)
    render_3d_napari(locdata_blobs_3d, viewer=viewer, bin_size=100, cmap='cyan', gamma=0.1, scale=(2, 2),
                     blending='additive')
    # napari.run()

    render_3d(locdata_blobs_3d, render_engine=RenderEngine.NAPARI)
    # napari.run()


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_3d_napari_empty(locdata_empty):
    render_3d_napari(locdata_empty, bin_size=100, cmap='viridis', gamma=0.1)
    napari.run()


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_3d_napari_single(locdata_single_localization, caplog):
    render_3d_napari(locdata_single_localization, bin_size=100, cmap='viridis', gamma=0.1)
    assert caplog.record_tuples[1] == ('locan.render.render3d', 30, 'Locdata carries a single localization.')
    napari.run()


def test_scatter_3d_mpl(locdata_3d):
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    scatter_3d_mpl(locdata_3d, ax=ax, text_kwargs=dict(color='r'), color='r')
    #plt.show()

    plt.close('all')


def test_scatter_3d_mpl_empty(locdata_empty):
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    scatter_3d_mpl(locdata_empty, ax=ax)
    #plt.show()

    plt.close('all')


def test_scatter_3d_mpl_single(locdata_single_localization, caplog):
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    scatter_3d_mpl(locdata_single_localization)
    assert caplog.record_tuples == [('locan.render.render3d', 30, 'Locdata carries a single localization.')]
    # plt.show()

    plt.close('all')


@pytest.mark.gui
@pytest.mark.parametrize("test_input, expected", list((member, 0) for member in list(RenderEngine)))
def test_render_3d(locdata_blobs_3d, test_input, expected):
    if HAS_DEPENDENCY["napari"] and test_input == RenderEngine.NAPARI:
        render_3d(locdata_blobs_3d, render_engine=test_input)
        # napari.run()
    else:
        with pytest.raises(NotImplementedError):
            render_3d(locdata_blobs_3d, render_engine=test_input)
    # plt.show()

    plt.close('all')


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_3d_rgb_napari(locdata_blobs_3d):
    locdata_0 = locdata_blobs_3d
    locdata_1 = transform_affine(locdata_blobs_3d, offset=(20, 0, 0))
    render_3d_rgb_napari([locdata_0, locdata_1], bin_size=20)

    napari.run()



"""
Localization data.

This module contains classes and functions to deal with single-molecule localization data.
All functions provide or modify LocData objects.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   locdata
   properties
   hulls
   rois
   region
   region_utils
   register
   aggregate
   filter
   transform
   cluster
   tracking
   metadata_utils

"""
from locan.data.aggregate import *
from locan.data.cluster import *
from locan.data.properties import *
from locan.data.transform import *
from locan.data.filter import *
from locan.data.hulls import *
from locan.data.locdata import *
from locan.data.region import *
from locan.data.region_utils import *
from locan.data.register import *
from locan.data.rois import *
from locan.data.tracking import *
from locan.data.metadata_utils import *


__all__ = []
__all__.extend(aggregate.__all__)
__all__.extend(cluster.__all__)
__all__.extend(properties.__all__)
__all__.extend(transform.__all__)
__all__.extend(filter.__all__)
__all__.extend(hulls.__all__)
__all__.extend(locdata.__all__)
__all__.extend(region.__all__)
__all__.extend(region_utils.__all__)
__all__.extend(register.__all__)
__all__.extend(rois.__all__)
__all__.extend(tracking.__all__)
__all__.extend(metadata_utils.__all__)



import pytest

import numpy as np
import matplotlib.pyplot as plt  # needed for visual inspection

from locan import regions_union, expand_region, surrounding_region
from locan import render_2d_mpl, scatter_2d_mpl  # needed for visual inspection
from locan import Region, Interval, Rectangle, Polygon, MultiPolygon, RoiRegion


def test_regions_union_Rectangles():
    region_0 = Rectangle((0, 0), 1, 1, 0)
    region_1 = Rectangle((2, 2), 1, 1, 0)
    regions = [region_0, region_1]
    union = regions_union(regions)
    assert isinstance(union, MultiPolygon)
    assert union.region_measure == 2

    region_0 = Rectangle((0, 0), 1, 1, 0)
    region_1 = Rectangle((1, 1), 1, 1, 0)
    region_2 = Rectangle((0, 0.5), 1, 1, 0)
    regions = [region_0, region_1, region_2]
    union = regions_union(regions)
    assert isinstance(union, Polygon)
    assert union.region_measure == 2.5

    region_0 = Polygon([(2, 2), (2, 3), (3, 3), (3, 2.5), (2, 2)])
    region_1 = Rectangle((10, 10), 1, 1, 0)
    regions = [region_0, region_1]
    union = regions_union(regions)
    assert isinstance(union, MultiPolygon)
    assert union.region_measure == 1.75

    # visualize
    fig, ax = plt.subplots()
    ax.scatter(*region_0.points.T)
    ax.add_patch(region_0.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(region_1.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(region_2.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(union.as_artist(fill=True, alpha=0.2, color='Red'))
    # plt.show()

    plt.close('all')


def test_extend_Interval():
    region = Interval(0, 1)
    expanded_region = expand_region(region, distance=1)
    assert expanded_region.region_measure == pytest.approx(3)
    assert expanded_region.bounds == (-1, 2)

    with pytest.raises(NotImplementedError):
        support = Interval(-1, 2)
        expanded_region_with_support = expand_region(region, distance=2, support=support)
        assert isinstance(expanded_region_with_support, Region)
        assert expanded_region_with_support.region_measure == 3
        assert expanded_region_with_support.bounds == (-1, 2)


def test_extend_Rectangle():
    region = Rectangle((0, 0), 1, 1, 0)
    extended_region = expand_region(region, distance=1)
    assert extended_region.region_measure == pytest.approx(8.13654849054594)

    region = Rectangle((0, 0), 2, 1, 0)
    support = Rectangle((0, 0), 3, 2, 0)
    extended_region = expand_region(region, distance=2)
    extended_region_with_support = expand_region(region, distance=2, support=support)
    assert isinstance(extended_region, Region)
    assert isinstance(extended_region_with_support, Region)
    assert region.region_measure == 2
    assert extended_region.region_measure == pytest.approx(26.546193962183754)
    assert extended_region_with_support.region_measure == 6

    # visualize
    fig, ax = plt.subplots()
    ax.scatter(*region.points.T)
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(extended_region.as_artist(fill=True, alpha=0.2, color='Red'))
    ax.add_patch(extended_region_with_support.as_artist(fill=True, alpha=0.2, color='Green'))
    # plt.show()

    plt.close('all')


def test_extend_Rectangles():
    region_0 = Rectangle((0, 0), 1, 1, 0)
    region_1 = Rectangle((2, 2), 1, 1, 0)
    region = regions_union([region_0, region_1])
    extended_region = expand_region(region, distance=0.1)
    assert isinstance(extended_region, Region)
    assert extended_region.region_measure == pytest.approx(2.8627309698109196)

    region_0 = Rectangle((0, 0), 1, 1, 0)
    region_1 = Rectangle((1, 1), 1, 1, 0)
    region_2 = Rectangle((0, 0.5), 1, 1, 0)
    region = regions_union([region_0, region_1, region_2])
    extended_region = expand_region(region, distance=2)
    assert isinstance(extended_region, Region)
    assert region_0.region_measure == 1
    assert extended_region.region_measure == pytest.approx(29.766960686847877)

    # visualize
    fig, ax = plt.subplots()
    ax.scatter(*region_0.points.T)
    ax.add_patch(region_0.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(region_1.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(region_2.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(extended_region.as_artist(fill=True, alpha=0.2, color='Red'))
    # plt.show()

    plt.close('all')


def test_extend_Polygon():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region = Polygon(points, holes)
    support = Rectangle((0, 0), 1, 1, 0)
    extended_region = expand_region(region, distance=0.1)
    extended_region_with_support = expand_region(region, distance=0.1, support=support)
    assert isinstance(extended_region, Region)
    assert isinstance(extended_region_with_support, Region)
    assert region.region_measure == pytest.approx(0.7350000000000001)
    assert extended_region.region_measure == pytest.approx(1.1431688585174895)
    assert extended_region_with_support.region_measure == pytest.approx(0.8493033988749895)

    # visualize
    fig, ax = plt.subplots()
    ax.scatter(*region.points.T)
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(extended_region.as_artist(fill=True, alpha=0.2, color='Red'))
    ax.add_patch(extended_region_with_support.as_artist(fill=True, alpha=0.2, color='Green'))
    # plt.show()

    plt.close('all')


def test_extend_MultiPolygon():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region = Polygon(points, holes)

    support = Rectangle((0, 0), 1, 1, 0)
    extended_region = expand_region(region, distance=0.1)
    extended_region_with_support = expand_region(region, distance=0.1, support=support)
    assert isinstance(extended_region, Region)
    assert isinstance(extended_region_with_support, Region)
    assert region.region_measure == pytest.approx(0.7350000000000001)
    assert extended_region.region_measure == pytest.approx(1.1431688585174895)
    assert extended_region_with_support.region_measure == pytest.approx(0.8493033988749895)

    # visualize
    fig, ax = plt.subplots()
    ax.scatter(*region.points.T)
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(extended_region.as_artist(fill=True, alpha=0.2, color='Red'))
    ax.add_patch(extended_region_with_support.as_artist(fill=True, alpha=0.2, color='Green'))
    # plt.show()

    plt.close('all')


def test_surrounding_region_Rectangles():
    region_0 = Rectangle((0, 0), 1, 1, 0)
    region_1 = Rectangle((1, 1), 1, 1, 0)
    region_2 = Rectangle((0, 0.5), 1, 1, 0)
    region = regions_union([region_0, region_1, region_2])
    extended_region = surrounding_region(region, distance=2)
    assert isinstance(extended_region, Region)
    assert region_0.region_measure == 1
    assert extended_region.region_measure == pytest.approx(27.266960686847888)

    # visualize
    fig, ax = plt.subplots()
    ax.scatter(*region_0.points.T)
    ax.add_patch(region_0.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(region_1.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(region_2.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(extended_region.as_artist(fill=True, alpha=0.2, color='Red'))
    # plt.show()

    plt.close('all')


def test_surrounding_region_Polygon():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region = Polygon(points, holes)
    support = Rectangle((0, 0), 1, 1, 0)
    extended_region = surrounding_region(region, distance=0.1)
    extended_region_with_support = surrounding_region(region, distance=0.1, support=support)
    assert isinstance(extended_region, Region)
    assert isinstance(extended_region_with_support, Region)
    assert region.region_measure == pytest.approx(0.7350000000000001)
    assert extended_region.region_measure == pytest.approx(0.4081688585174893)
    assert extended_region_with_support.region_measure == pytest.approx(0.11430339887498947)

    # visualize
    fig, ax = plt.subplots()
    ax.scatter(*region.points.T)
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.add_patch(extended_region.as_artist(fill=True, alpha=0.2, color='Red'))
    ax.add_patch(extended_region_with_support.as_artist(fill=True, alpha=0.2, color='Green'))
    # plt.show()

    plt.close('all')


def test_extend_region_RoiRegion():
    rr = RoiRegion(region_type='rectangle', region_specs=((0, 0), 1, 1, 0))
    er = expand_region(region=rr, distance=1, support=None)
    assert rr.region_measure == pytest.approx(1)

    # visualize
    fig, ax = plt.subplots()
    ax.scatter(*rr.points.T)
    ax.add_patch(rr.as_artist(fill=True))
    ax.scatter(np.array(er.exterior.coords)[:, 0], np.array(er.exterior.coords)[:, 1])
    # plt.show()

    assert er.area == pytest.approx(8.13654849054594)
    er = expand_region(region=rr.to_shapely(), distance=1, support=None)
    assert er.area == pytest.approx(8.13654849054594)

    rr = RoiRegion(region_type='ellipse', region_specs=((0, 0), 1, 3, 0))
    sr = expand_region(region=rr, distance=1, support=None)
    assert sr.area == pytest.approx(12.169350575721547)

    rr = RoiRegion(region_type='polygon', region_specs=((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0)))
    sr = expand_region(region=rr, distance=1, support=None)
    assert sr.area == pytest.approx(7.50457995299987)

    rr = RoiRegion(region_type='polygon', region_specs=((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0)))
    sup = RoiRegion(region_type='rectangle', region_specs=((0, 0), 2, 2, 0))
    sr = expand_region(region=rr, distance=1, support=sup)
    assert sr.area == pytest.approx(3.762009021216654)

    rr = RoiRegion(region_type='rectangle', region_specs=((0, 0), 1, 1, 0))
    sr = expand_region(region=rr, distance=1, support=None)
    assert sr.area == pytest.approx(8.13654849054594)
    assert isinstance(sr, Polygon)
    rr_1 = RoiRegion(region_type='rectangle', region_specs=((10, 10), 1, 1, 0))
    region = regions_union([rr, rr_1])
    sr = expand_region(region=region, distance=1, support=None)
    assert sr.area == pytest.approx(2*8.13654849054594)

    # visualize
    fig, ax = plt.subplots()
    ax.scatter((-1, 3), (-1, 3))
    ax.add_patch(rr.as_artist(fill=True))
    ax.add_patch(sr.as_artist(fill=True))
    # plt.show()

    # visualize
    # fig, ax = plt.subplots()
    # #ax.scatter((-1, 3), (-1, 3))
    # ax.add_patch(rr.as_artist(fill=True))
    # ax.add_patch(rr_1.as_artist(fill=True))
    # sr_region = RoiRegion.from_shapely(region='shapelyMultiPolygon', shapely_obj=sr)
    # for pat in sr_region.as_artist(fill=False):
    #     ax.add_patch(pat)
    # plt.show()

    plt.close('all')


def test_surrounding_region():
    rr = RoiRegion(region_type='rectangle', region_specs=((0, 0), 1, 1, 0))
    sr = surrounding_region(region=rr, distance=1, support=None)
    assert sr.area == pytest.approx(7.136548490545939)

    rr = RoiRegion(region_type='ellipse', region_specs=((0, 0), 1, 3, 0))
    sr = surrounding_region(region=rr, distance=1, support=None)
    assert sr.area == pytest.approx(9.816939207812093)

    rr = RoiRegion(region_type='polygon', region_specs=((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0)))
    sr = surrounding_region(region=rr, distance=1, support=None)
    assert sr.area == pytest.approx(6.754579952999869)

    rr = RoiRegion(region_type='polygon', region_specs=((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0)))
    sup = RoiRegion(region_type='rectangle', region_specs=((0, 0), 2, 2, 0))
    sr = surrounding_region(region=rr, distance=1, support=sup)
    assert sr.area == pytest.approx(3.012009021216654)

    rr = RoiRegion(region_type='rectangle', region_specs=((0, 0), 1, 1, 0))
    rr_1 = RoiRegion(region_type='rectangle', region_specs=((0, 0), 1, 1, 0))
    sr = surrounding_region(region=regions_union([rr, rr_1]), distance=1, support=None)
    assert sr.area == pytest.approx(7.136548490545939)
    assert isinstance(sr, Polygon)
    rr_1 = RoiRegion(region_type='rectangle', region_specs=((10, 10), 1, 1, 0))
    sr = surrounding_region(region=regions_union([rr, rr_1]), distance=1, support=None)
    assert sr.area == pytest.approx(2*7.136548490545939)
    assert isinstance(sr, MultiPolygon)

    # visualize
    # fig, ax = plt.subplots()
    # #ax.scatter((-1, 3), (-1, 3))
    # ax.add_patch(rr.as_artist(fill=True))
    # ax.add_patch(rr_1.as_artist(fill=True))
    # sr_region = RoiRegion.from_shapely(region='shapelyMultiPolygon', shapely_obj=sr)
    # for pat in sr_region.as_artist(fill=False):
    #     ax.add_patch(pat)
    # plt.show()

    plt.close('all')



import pytest

from locan import ROOT_DIR
from locan.data import metadata_pb2


@pytest.fixture
def metadata_pb2_Metadata_v0p11():
    """
    Fixture for protobuf message metadata_pb2.Metadata in locan <= 0.11
    """
    metadata = metadata_pb2.Metadata()

    metadata.identifier = "111"
    metadata.comment = "comment"
    metadata.creation_date = "1111-11-11 11:11:11 +0100"
    metadata.modification_date = "1111-11-11 11:11:11 +0100"
    metadata.production_date = "1111-11-11 11:11:11 +0100"
    metadata.source = 0
    metadata.state = 0
    metadata.history.add(name="function_name", parameter="parameters")
    metadata.ancestor_identifiers.append("110")
    metadata.unit.add(property="name", unit="s")
    metadata.unit.add(property="name_2", unit="m")
    metadata.element_count = 111
    metadata.frame_count = 111
    metadata.file_type = 0
    metadata.file_path = "file_path"
    metadata.experimental_setup["test_key"] = "setup"
    metadata.experimental_sample["test_key"] = "sample"
    metadata.map["test_key"] = "map_value"
    metadata.map["test_key_2"] = "map_value_2"

    return metadata


@pytest.fixture
def metadata_pb2_Metadata_v0p11_latest():
    """
    Fixture for protobuf message metadata_pb2.Metadata in locan <= 0.11
    as usable in latest protobuf definition.
    """
    metadata = metadata_pb2.Metadata()

    metadata.identifier = "111"
    metadata.comment = "comment"
    # metadata.creation_date = "1111-11-11 11:11:11 +0100"
    # metadata.modification_date = "1111-11-11 11:11:11 +0100"
    # metadata.production_date = "1111-11-11 11:11:11 +0100"
    metadata.source = 0
    metadata.state = 0
    metadata.history.add(name="function_name", parameter="parameters")
    metadata.ancestor_identifiers.append("110")
    # metadata.unit.add(property="name", unit="s")
    # metadata.unit.add(property="name_2", unit="m")
    metadata.element_count = 111
    metadata.frame_count = 111
    # metadata.file_type = 0
    # metadata.file_path = "file_path"
    # metadata.experimental_setup["test_key"] = "setup"
    # metadata.experimental_sample["test_key"] = "sample"
    metadata.map["test_key"] = "map_value"
    metadata.map["test_key_2"] = "map_value_2"

    return metadata


@pytest.fixture
def metadata_pb2_Metadata_v0p12():
    """
    Fixture for protobuf message metadata_pb2.Metadata in locan >= 0.12
    """
    operation = metadata_pb2.Operation()

    operation.name = "function_name"
    operation.parameter = "parameters"

    file = metadata_pb2.File()

    file.identifier = "111"
    file.comment = "comment"
    file.type = 0
    file.path = "file_path"
    file.groups.append("group 1")

    address = metadata_pb2.Address()

    address.address_lines.append("address_line")
    address.city = "city"
    address.city_code = "city_code"
    address.country = "country"

    affiliation = metadata_pb2.Affiliation()

    affiliation.institute = "institute"
    affiliation.department = "department"
    affiliation.address.CopyFrom(address)

    person = metadata_pb2.Person()

    person.identifier = "111"
    person.comment = "comment"
    person.first_name = "first_name"
    person.last_name = "last_name"
    person.title = "title"
    person.affiliations.append(affiliation)
    person.address.CopyFrom(address)
    person.emails.append("email_1")
    person.roles.append("roles_1")

    experimental_sample = metadata_pb2.ExperimentalSample()

    experimental_sample.identifier = "111"
    experimental_sample.comment = "comment"
    experimental_sample.targets.append("targets")
    experimental_sample.fluorophores.append("fluorophores")
    experimental_sample.buffers.append("buffers")
    experimental_sample.map["test_key"] = "map_value"

    illumination = metadata_pb2.Illumination()

    illumination.identifier = "111"
    illumination.comment = "comment"
    illumination.lightsource = "lightsource"
    illumination.power = 1.1
    illumination.area = 1.1
    illumination.power_density = 1.1
    illumination.wavelength = 1.1
    illumination.map["test_key"] = "map_value"

    camera = metadata_pb2.Camera()

    camera.identifier = "111"
    camera.comment = "comment"
    camera.name = "name"
    camera.model = "model"
    camera.gain = 1.1
    camera.electrons_per_count = 1.1
    camera.integration_time.FromMilliseconds(10)
    camera.pixel_count_x = 1
    camera.pixel_count_y = 1
    camera.pixel_size_x = 1.1
    camera.pixel_size_y = 1.1
    camera.flipped = True
    camera.map["test_key"] = "map_value"

    detection = metadata_pb2.Detection()

    detection.identifier = "111"
    detection.comment = "comment"
    detection.camera.CopyFrom(camera)
    detection.map["test_key"] = "map_value"

    acquisition = metadata_pb2.Acquisition()

    acquisition.identifier = "111"
    acquisition.comment = "comment"
    acquisition.frame_count = 1
    acquisition.frame_of_interest_first = 1
    acquisition.frame_of_interest_last = 2
    acquisition.time_start.FromJsonString("1970-01-01T00:00:00Z")
    acquisition.time_end.FromJsonString("1970-01-01T00:01:00Z")
    acquisition.stack_count = 1
    acquisition.stack_step_count = 1
    acquisition.stack_step_size = 1.1
    acquisition.map["test_key"] = "map_value"

    lightsheet = metadata_pb2.Lightsheet()

    lightsheet.identifier = "111"
    lightsheet.comment = "comment"
    lightsheet.angle_x = 1.1
    lightsheet.angle_y = 1.1
    lightsheet.angle_z = 1.1
    lightsheet.map["test_key"] = "map_value"

    optical_unit = metadata_pb2.OpticalUnit()

    optical_unit.identifier = "111"
    optical_unit.comment = "comment"
    optical_unit.illumination.CopyFrom(illumination)
    optical_unit.detection.CopyFrom(detection)
    optical_unit.acquisition.CopyFrom(acquisition)
    optical_unit.lightsheet.CopyFrom(lightsheet)

    experimental_setup = metadata_pb2.ExperimentalSetup()

    experimental_setup.identifier = "111"
    experimental_setup.comment = "comment"
    experimental_setup.optical_units.append(optical_unit)
    experimental_setup.map["test_key"] = "map_value"

    experiment = metadata_pb2.Experiment()

    experiment.identifier = "111"
    experiment.comment = "comment"
    experiment.experimenters.append(person)
    experiment.samples.append(experimental_sample)
    experiment.setups.append(experimental_setup)
    experiment.map["test_key"] = "map_value"

    localizer = metadata_pb2.Localizer()

    localizer.identifier = "111"
    localizer.comment = "comment"
    localizer.software = "software"
    localizer.intensity_threshold = 1.1
    localizer.psf_fixed = True
    localizer.psf_size = 1.1
    localizer.map["test_key"] = "map_value"

    relation = metadata_pb2.Relation()

    relation.identifier = "111"
    relation.comment = "comment"
    relation.file.CopyFrom(file)
    relation.map["test_key"] = "map_value"

    metadata = metadata_pb2.Metadata()

    metadata.identifier = "111"
    metadata.comment = "comment"
    metadata.creation_time.FromJsonString("1111-11-11T11:11:11+01:00")
    metadata.modification_time.FromJsonString("1111-11-11T11:11:11+01:00")
    metadata.production_time.FromJsonString("1111-11-11T11:11:11+01:00")
    metadata.source = 0
    metadata.state = 0
    metadata.history.append(operation)
    metadata.ancestor_identifiers.append("110")
    metadata.element_count = 1
    metadata.frame_count = 1
    metadata.properties.add(name='position_x', unit='nm', type='float')
    metadata.localization_properties.add(name='position_x', unit='nm', type='float')
    metadata.file.CopyFrom(file)
    metadata.relations.append(relation)
    metadata.experiment.CopyFrom(experiment)
    metadata.localizer.CopyFrom(localizer)
    metadata.map["test_key"] = "map_value"

    return metadata


@pytest.mark.skip("Use only to write new file with protobuf message!")
def test_save_proto_message_locdata_metadata(metadata_pb2_Metadata_v0p12):
    """
    Write only to produce new file with metadata message. Keep old files to test new proto definitions.
    Do NOT overwrite the old versions:
    1) "tests/test_data/protobuf_message_metadata_pb2.Metadata_v0p11"
    2) "tests/test_data/protobuf_message_metadata_pb2.Metadata_v0p12"
    """
    metadata = metadata_pb2_Metadata_v0p12

    print(metadata)

    path = ROOT_DIR / "tests/test_data/protobuf_message_metadata_pb2.Metadata_v0px"
    with open(path, "wb") as file:
        file.write(metadata.SerializeToString())

    metadata_new = metadata_pb2.Metadata()
    with open(path, "rb") as file:
        metadata_new.ParseFromString(file.read())

    assert metadata_new == metadata


def test_read_proto_message_locdata_metadata_v0p11(metadata_pb2_Metadata_v0p11_latest):
    metadata = metadata_pb2_Metadata_v0p11_latest
    metadata_new = metadata_pb2.Metadata()
    path = ROOT_DIR / "tests/test_data/protobuf_message_metadata_pb2.Metadata_v0p11"
    with open(path, "rb") as file:
        metadata_new.ParseFromString(file.read())

    for key in metadata_new.DESCRIPTOR.fields_by_name.keys():
        assert key in metadata.DESCRIPTOR.fields_by_name.keys()
        if key == "map":
            assert dict(getattr(metadata_new, key)) == dict(getattr(metadata, key))
        else:
            assert getattr(metadata_new, key) == getattr(metadata, key)


def test_read_proto_message_locdata_metadata(metadata_pb2_Metadata_v0p12):
    metadata = metadata_pb2_Metadata_v0p12

    # print(metadata)

    path = ROOT_DIR / "tests/test_data/protobuf_message_metadata_pb2.Metadata_v0p12"
    metadata_new = metadata_pb2.Metadata()
    with open(path, "rb") as file:
        metadata_new.ParseFromString(file.read())

    # print(metadata_new)

    for key in metadata_new.DESCRIPTOR.fields_by_name.keys():
        assert key in metadata.DESCRIPTOR.fields_by_name.keys()

    assert metadata_new == metadata



import locan
import pytest
import numpy as np
import matplotlib.pyplot as plt  # this import is needed for visual inspection
from lmfit.model import ModelResult
from lmfit.models import ConstantModel, LinearModel, PolynomialModel

from locan import LocData
from locan.dependencies import HAS_DEPENDENCY
from locan import Drift, DriftComponent
from locan.analysis.drift import _LmfitModelFacade, _ConstantModelFacade, _ConstantZeroModelFacade, \
    _ConstantOneModelFacade, _SplineModelFacade, _estimate_drift_cc, _estimate_drift_icp


# data to evaluate fitting
x = np.array([1, 2, 4, 6, 9, 10, 11, 15, 16, 20])
y = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])


def test__LmfitModelFacade():
    model = _LmfitModelFacade(LinearModel())
    model.fit(x=x, y=y, verbose=True)
    assert model.eval(5) == pytest.approx(3.401234567901236)
    assert len(model.eval([2, 5])) == 2
    model.plot()
    # plt.show()

    plt.close('all')


def test__ConstantModelFacade():
    model = _ConstantModelFacade()
    model.fit(x=x, y=y, verbose=True)
    assert model.eval(5) == pytest.approx(5.500000000011998)
    assert len(model.eval([2, 5])) == 2
    model.plot()
    # plt.show()

    plt.close('all')


def test__ConstantZeroModelFacade():
    model = _ConstantZeroModelFacade()
    model.fit(x=x, y=y, verbose=True)
    assert model.eval(5) == 0
    assert len(model.eval([2, 5])) == 2
    model.plot()
    # plt.show()

    plt.close('all')


def test__ConstantOneModelFacade():
    model = _ConstantOneModelFacade()
    model.fit(x=x, y=y, verbose=True)
    assert model.eval(5) == 1
    assert len(model.eval([2, 5])) == 2
    model.plot()
    # plt.show()

    plt.close('all')


def test__SplineModelFacade():
    model = _SplineModelFacade()
    model.fit(x=x, y=y, verbose=True)
    # print(model.eval(5))
    # print(model.eval((3, 5)))
    # print(model.eval([3, 5]))
    # print(model.eval(np.array([3, 5])))
    assert model.eval(5) == pytest.approx(3.417849958061275)
    assert len(model.eval([2, 5])) == 2
    # model.plot()
    model.fit(x=x, y=y, s=0)
    assert model.eval(5) == pytest.approx(3.5332295857553917)
    # model.plot()
    # plt.show()
    model = _SplineModelFacade(s=0)
    model.fit(x=x, y=y, verbose=True)
    assert model.eval(5) == pytest.approx(3.5332295857553917)


def test_DriftComponent():
    drift_component = DriftComponent(type='linear')
    assert drift_component.model_result is None
    with pytest.raises(AttributeError):
        drift_component.eval(3)

    with pytest.raises(TypeError):
        DriftComponent(type='undefined')

    drift_component = DriftComponent(type='linear').fit(x, y, verbose=True)
    assert isinstance(drift_component.model.model, LinearModel)
    assert isinstance(drift_component.model_result, ModelResult)
    assert drift_component.eval(4) == pytest.approx(2.924242424242425)
    assert len(drift_component.eval([1, 4]) == 2)

    drift_component = DriftComponent(type='zero').fit(x, y, verbose=True)
    assert isinstance(drift_component.model.model, ConstantModel)
    assert isinstance(drift_component.model_result, ModelResult)
    assert drift_component.model_result.best_fit == 0
    assert drift_component.eval(4) == 0
    assert len(drift_component.eval([1, 4]) == 2)

    drift_component = DriftComponent(type='one').fit(x, y, verbose=True)
    assert isinstance(drift_component.model.model, ConstantModel)
    assert isinstance(drift_component.model_result, ModelResult)
    assert drift_component.model_result.best_fit == 1
    assert drift_component.eval(4) == 1
    assert len(drift_component.eval([1, 4]) == 2)

    drift_component = DriftComponent(type='constant').fit(x, y, verbose=True)
    assert isinstance(drift_component.model.model, ConstantModel)
    assert isinstance(drift_component.model_result, ModelResult)
    assert drift_component.model_result.best_fit == pytest.approx(5.500000000011998)
    assert drift_component.eval(4) == pytest.approx(5.500000000011998)
    assert len(drift_component.eval([1, 4]) == 2)

    drift_component = DriftComponent(type='polynomial', degree=2).fit(x, y, verbose=True)
    assert isinstance(drift_component.model.model, PolynomialModel)
    assert isinstance(drift_component.model_result, ModelResult)

    drift_component = DriftComponent(type='spline').fit(x, y, verbose=True, s=0.3)
    assert drift_component.eval(4) == pytest.approx(3.0972390587725087)
    assert len(drift_component.eval([1, 4]) == 2)

    drift_component = DriftComponent(type=LinearModel()).fit(x, y, verbose=True)
    assert isinstance(drift_component.model.model, LinearModel)
    assert isinstance(drift_component.model_result, ModelResult)
    assert drift_component.eval(4) == pytest.approx(2.924242424242425)
    assert len(drift_component.eval([1, 4]) == 2)
    # plt.show()

    plt.close('all')


def test_Drift_empty(caplog):
    drift = Drift().compute(LocData())
    drift.plot()
    drift.apply_correction()
    drift.fit_transformation()
    drift.fit_transformations()
    assert caplog.record_tuples == [('locan.analysis.drift', 30, 'Locdata is empty.'),
                                    ('locan.analysis.drift', 30, 'No transformations available to be applied.'),
                                    ('locan.analysis.drift', 30, 'No transformations available to be fitted.'),
                                    ('locan.analysis.drift', 30, 'No transformations available to be fitted.')]




def test__estimate_drift_cc(locdata_blobs_2d):
    # locan.render_2d_mpl(locdata_blobs_2d, bin_size=10, rescale=locan.Trafo.EQUALIZE)
    # plt.show()

    collection, transformations = _estimate_drift_cc(locdata_blobs_2d,
                                                     chunks=None, chunk_size=25,
                                                     target='first', bin_size=10,
                                                     kwargs_chunk=None, kwargs_register=None)
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2

    index_groups = (range(0, 25), range(25, 50))
    collection, transformations = _estimate_drift_cc(locdata_blobs_2d,
                                                     chunks=index_groups, chunk_size=None,
                                                     target='first', bin_size=10,
                                                     kwargs_chunk=None, kwargs_register=None)
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2

    collection, transformations = _estimate_drift_cc(locdata_blobs_2d,
                                                     chunks=None, chunk_size=25,
                                                     target='previous', bin_size=10,
                                                     kwargs_chunk=None, kwargs_register=None)
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test__estimate_drift_icp(locdata_blobs_2d):
    collection, transformations = _estimate_drift_icp(locdata_blobs_2d,
                                                     chunks=None, chunk_size=25,
                                                     target='first',
                                                     kwargs_chunk=None, kwargs_register=None)
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2

    index_groups = (range(0, 25), range(25, 50))
    collection, transformations = _estimate_drift_icp(locdata_blobs_2d,
                                                     chunks=index_groups, chunk_size=None,
                                                     target='first',
                                                     kwargs_chunk=None, kwargs_register=None)
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2

    collection, transformations = _estimate_drift_icp(locdata_blobs_2d,
                                                     chunks=None, chunk_size=25,
                                                     target='previous',
                                                     kwargs_chunk=None, kwargs_register=None)
    assert len(collection) == 2
    assert isinstance(collection, LocData)
    assert len(transformations) == 2
    assert len(transformations[0].matrix) == 2
    assert len(transformations[0].offset) == 2


def test_Drift(locdata_blobs_2d):
    drift = Drift(chunk_size=15, target='first', method='cc').compute(locdata_blobs_2d)
    assert isinstance(drift.locdata, LocData)
    assert isinstance(drift.collection, LocData)
    assert drift.transformations[0]._fields == ('matrix', 'offset')
    assert len(drift.transformations) == 4
    assert drift.transformation_models['matrix'] is None
    assert drift.transformation_models['offset'] is None

    std_trafo_models = drift._transformation_models_for_identity_matrix()
    assert std_trafo_models['matrix'][0].type == 'one'
    std_trafo_models = drift._transformation_models_for_zero_offset()
    assert std_trafo_models['offset'][0].type == 'zero'

    assert drift.transformation_models['matrix'] is None
    drift.fit_transformation(slice_data=slice(None), transformation_component='matrix', element=1,
                             drift_model='linear', verbose=False)
    assert isinstance(drift.transformation_models['matrix'][0], DriftComponent)
    assert drift.transformation_models['matrix'][0].type == 'one'
    assert drift.transformation_models['matrix'][1].type == 'linear'
    assert drift.transformation_models['matrix'][2].type == 'zero'

    assert drift.transformation_models['offset'] is None
    drift.fit_transformation(slice_data=slice(None), transformation_component='offset', element=1,
                             drift_model='linear', verbose=False)
    assert drift.transformation_models['offset'][0].type == 'zero'
    assert drift.transformation_models['offset'][1].type == 'linear'

    drift.fit_transformation(slice_data=slice(None), transformation_component='matrix', element=1,
                             drift_model=None, verbose=False)
    assert drift.transformation_models['matrix'][1].type == 'linear'

    drift.fit_transformations(matrix_models=None, offset_models=None, verbose=False)
    assert drift.transformation_models['matrix'] is None
    assert drift.transformation_models['offset'] is None

    drift.fit_transformations(slice_data=slice(None),
                              matrix_models=('constant', 'linear', DriftComponent(type='polynomial', degree=2), None),
                              offset_models=('spline', DriftComponent(type='spline', s=100)),
                              verbose=False)
    assert drift.transformation_models['matrix'][0].type == 'constant'
    assert drift.transformation_models['matrix'][1].type == 'linear'
    assert drift.transformation_models['matrix'][2].type == 'polynomial'
    assert drift.transformation_models['matrix'][3].type == 'one'
    assert drift.transformation_models['offset'][0].type == 'spline'
    assert drift.transformation_models['offset'][1].type == 'spline'

    assert drift.locdata_corrected is None
    new_locdata = drift._apply_correction_on_chunks()
    assert len(new_locdata) == len(locdata_blobs_2d)

    drift.transformation_models['matrix'] = None
    drift.transformation_models['offset'] = None
    with pytest.raises(AttributeError):
        drift._apply_correction_from_model(drift.locdata_rapidSTORM_2d)

    drift.fit_transformation(slice_data=slice(None), transformation_component='matrix', element=1,
                             drift_model='linear', verbose=True)
    assert drift.transformation_models['matrix'][1].type == 'linear'
    # print(drift.transformation_models['matrix'][1].model)
    # print(drift.transformation_models['matrix'][1].model_result)
    # print(drift.transformation_models['matrix'][1].eval(3))

    drift.fit_transformation(slice_data=slice(None), transformation_component='matrix', element=0,
                             drift_model='one', verbose=True)
    assert drift.transformation_models['matrix'][0].type == 'one'
    # print(drift.transformation_models['matrix'][0].model)
    # print(drift.transformation_models['matrix'][0].model_result)
    # print(drift.transformation_models['matrix'][0].eval(3))

    drift.fit_transformations(slice_data=slice(None),
                              matrix_models=['linear'] * 4,
                              offset_models=['linear'] * 2,
                              verbose=False)
    transformed_points = drift._apply_correction_from_model(drift.locdata)
    assert len(transformed_points) == len(locdata_blobs_2d)

    drift.apply_correction(from_model=True)
    assert isinstance(drift.locdata_corrected, LocData)
    drift.apply_correction(from_model=False)
    assert isinstance(drift.locdata_corrected, LocData)
    drift.apply_correction(locdata=drift.collection.references[0], from_model=True)
    assert isinstance(drift.locdata_corrected, LocData)
    assert len(drift.locdata_corrected) == len(drift.collection.references[0])

    # change target and fit_transformations
    drift = Drift(chunk_size=25, target='previous', method='cc').compute(locdata_blobs_2d)
    assert isinstance(drift.collection, LocData)
    assert drift.transformations[0]._fields == ('matrix', 'offset')
    assert len(drift.transformations) == 2

    drift.plot(transformation_component='matrix', element=None)
    drift.plot(transformation_component='matrix', element=3)
    drift.plot(transformation_component='offset', element=None)
    drift.plot(transformation_component='offset', element=0)
    # plt.show()

    with pytest.raises(NotImplementedError):
        drift.fit_transformations(slice_data=slice(None),
                                  matrix_models=['linear'] * 4,
                                  offset_models=['linear'] * 2,
                                  verbose=False)

    plt.close('all')


def test_Drift_chain(locdata_blobs_2d):
    # apply chained functions
    drift = Drift(chunk_size=25, target='first', method='cc').\
        compute(locdata_blobs_2d).\
        fit_transformations(slice_data=slice(0, 2),
                            matrix_models=None,
                            offset_models=(None, LinearModel())).\
        apply_correction()
    assert len(drift.locdata_corrected) == len(drift.locdata)
    assert isinstance(drift.locdata, LocData)
    assert isinstance(drift.collection, LocData)
    assert drift.transformations[0]._fields == ('matrix', 'offset')
    assert len(drift.transformations) == 2


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test_Drift_with_icp(locdata_blobs_2d):
    drift = Drift(chunk_size=15, target='first', method='icp').compute(locdata_blobs_2d)
    assert isinstance(drift.locdata, LocData)
    assert isinstance(drift.collection, LocData)
    assert drift.transformations[0]._fields == ('matrix', 'offset')
    assert len(drift.transformations) == 4
    assert drift.transformation_models['matrix'] is None
    assert drift.transformation_models['offset'] is None

    assert drift.transformation_models['matrix'] is None
    drift.fit_transformation(slice_data=slice(None), transformation_component='matrix', element=1,
                             drift_model='linear', verbose=False)
    assert isinstance(drift.transformation_models['matrix'][0], DriftComponent)
    assert drift.transformation_models['matrix'][0].type == 'one'
    assert drift.transformation_models['matrix'][1].type == 'linear'
    assert drift.transformation_models['matrix'][2].type == 'zero'

    assert drift.transformation_models['offset'] is None
    drift.fit_transformation(slice_data=slice(None), transformation_component='offset', element=1,
                             drift_model='linear', verbose=False)
    assert drift.transformation_models['offset'][0].type == 'zero'
    assert drift.transformation_models['offset'][1].type == 'linear'



import pytest
import numpy as np
from scipy.spatial import Delaunay

from locan.data.hulls.alpha_shape_2d import _circumcircle, _half_distance


def test__circumcircle_2d(locdata_2d):
    points = np.array([(0, 0), (1, 1+np.sqrt(2)), (1+np.sqrt(2), 1)])
    center, radius = _circumcircle(points, [2, 1, 0])
    assert radius == np.sqrt(2)
    assert np.array_equal(center, [1, 1])

    points = locdata_2d.coordinates
    triangulation = Delaunay(points)
    center, radius = _circumcircle(points, triangulation.simplices[0])
    assert radius == pytest.approx(1.8210786221487993)
    assert center[0] == pytest.approx(3.357142857142857)


def test__half_distance():
    points = np.array([(0, 0), (0, 1)])
    radius = _half_distance(points)
    assert radius == 0.5



"""

Functions for user interaction with paths and file names.

"""
from locan.dependencies import QT_BINDINGS, QtBindings

if QT_BINDINGS == QtBindings.PYSIDE2:
    from PySide2.QtWidgets import QApplication, QFileDialog
elif QT_BINDINGS == QtBindings.PYQT5:
    from PyQt5.QtWidgets import QApplication, QFileDialog

__all__ = ['file_dialog']


def file_dialog(directory=None, message='Select a file...', filter='Text files (*.txt);; All files (*)'):
    """
    Select file names in a ui dialog.

    Parameters
    ----------
    directory : str or None
        directory path to start dialog in. If None the current directory is used.

    message : str
        Hint what to do

    filter : str
        filter for file type

    Returns
    -------
    list of str
        list with file names or empty list
    """
    if QT_BINDINGS == QtBindings.NONE:
        raise ImportError('Function requires either PySide2 or PyQt5.')

    if directory is None:
        directory_ = './'
    else:
        directory_ = str(directory)

    if QT_BINDINGS == QtBindings.PYSIDE2:
        app = QApplication.instance()  # this is needed if the function is called twice in a row.
        if app is None:
            app = QApplication([])  # todo: [directory_] is not working - please fix!
    elif QT_BINDINGS == QtBindings.PYQT5:
        app = QApplication([directory_])

    fname = QFileDialog.getOpenFileNames(None, message, directory=directory_, filter=filter)

    if isinstance(fname, tuple):
        return fname[0]
    else:
        return str(fname)



"""

File input/output for localization data in DECODE files.

"""
import logging

import pandas as pd

from locan.dependencies import HAS_DEPENDENCY, needs_package
from locan.data.locdata import LocData
import locan.constants
from locan.data import metadata_pb2
from locan.locan_io.locdata.utilities import convert_property_types, convert_property_names

if HAS_DEPENDENCY["h5py"]: import h5py


__all__ = ['load_decode_header', 'load_decode_file']

logger = logging.getLogger(__name__)


@needs_package("h5py")
def _read_decode_header(file):
    """
    Read header from a DECODE single-molecule localization file and identify column names.

    Parameters
    ----------
    file : file-like
        File to read.

    Returns
    -------
    Tuple[list[str], Dict, Dict]
        Tuple with identifiers, meta and decode sections.
        Identifiers are list of valid dataset property keys as derived from the DECODE identifiers.
    """
    meta_data = dict(file['meta'].attrs)
    meta_decode = dict(file['decode'].attrs)

    # list identifiers
    identifiers = list(file['data'].keys())

    column_keys = []
    for i in identifiers:
        if i == "xyz":
            column_keys.extend(["position_x", "position_y", "position_z"])
        elif i == "xyz_cr":
            column_keys.extend(["x_cr", "y_cr", "z_cr"])
        elif i == "xyz_sig":
            column_keys.extend(["x_sig", "y_sig", "z_sig"])
        else:
            column_keys.append(i)

    column_keys = convert_property_names(properties=column_keys, property_mapping=locan.constants.DECODE_KEYS)

    return column_keys, meta_data, meta_decode


@needs_package("h5py")
def load_decode_header(path):
    """
    Load header from a DECODE single-molecule localization file and identify column names.

    The hdf5 file should contain the following keys: <KeysViewHDF5 ['data', 'decode', 'meta']>

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a DECODE file to load.

    Returns
    -------
    Tuple[list[str], Dict, Dict]
        Tuple with identifiers, meta and decode sections.
        Identifiers are list of valid dataset property keys as derived from the DECODE identifiers.
    """
    with h5py.File(path, 'r') as file:
        return _read_decode_header(file)


@needs_package("h5py")
def load_decode_file(path, nrows=None, convert=True):
    """
    Load data from a DECODE single-molecule localization file.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a Thunderstorm file to load.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.
    convert : bool
        If True convert types by applying type specifications in locan.constants.PROPERTY_KEYS.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with h5py.File(path, 'r') as file:
        columns, meta, decode = _read_decode_header(file)

        if file['data']['xyz'].shape == (0, 3):  # empty file
            logger.warning(f'File does not contain any data.')
            locdata = LocData()

        else:  # file not empty
            data = {}
            for key, value in file['data'].items():
                if value.shape is not None:
                    if key == "xyz":
                        for i, property_ in enumerate(['position_x', 'position_y', 'position_z']):
                            data[property_] = value[:nrows, i]
                    elif key == "xyz_cr":
                        for i, property_ in enumerate(["x_cr", "y_cr", "z_cr"]):
                            data[property_] = value[:nrows, i]
                    elif key == "xyz_sig":
                        for i, property_ in enumerate(["x_sig", "y_sig", "z_sig"]):
                            data[property_] = value[:nrows, i]
                    elif key in locan.constants.DECODE_KEYS:
                        data[locan.constants.DECODE_KEYS[key]] = value[:nrows]
                    else:
                        data[key] = value[:nrows]

            dataframe = pd.DataFrame(data)

            if convert:
                dataframe = convert_property_types(dataframe, types=locan.constants.PROPERTY_KEYS)

            locdata = LocData.from_dataframe(dataframe=dataframe)

    locdata.meta.source = metadata_pb2.EXPERIMENT
    locdata.meta.state = metadata_pb2.RAW
    locdata.meta.file.type = metadata_pb2.DECODE
    locdata.meta.file.path = str(path)

    for property_ in sorted(list(set(columns).intersection({'position_x', 'position_y', 'position_z'}))):
        locdata.meta.localization_properties.add(name=property_, unit=meta['xy_unit'], type="float")

    del locdata.meta.history[:]
    locdata.meta.history.add(name='load_decode_file', parameter='path={}, nrows={}'.format(path, nrows))

    return locdata



"""

File input/output for localization data.

There are functions for reading the following file structures
(with an indicator string in parenthesis - see also `locan.constants.FileType`):

* custom text file (CUSTOM)
* rapidSTORM file format (RAPIDSTORM) [1]_
* Elyra file format (ELYRA)
* Thunderstorm file format (THUNDERSTORM) [2]_
* asdf file format (ASDF) [3]_
* Nanoimager file format (NANOIMAGER)
* rapidSTORM track file format (RAPIDSTORMTRACK) [1]_
* smlm file format (SMLM) [4]_, [5]_
* decode file (DECODE) [6]_
* smap file format (SMAP) [7]_


References
----------
.. [1] Wolter S, Löschberger A, Holm T, Aufmkolk S, Dabauvalle MC, van de Linde S, Sauer M.,
   rapidSTORM: accurate, fast open-source software for localization microscopy.
   Nat Methods 9(11):1040-1, 2012, doi: 10.1038/nmeth.2224

.. [2] M. Ovesný, P. Křížek, J. Borkovec, Z. Švindrych, G. M. Hagen.
   ThunderSTORM: a comprehensive ImageJ plugin for PALM and STORM data analysis and super-resolution imaging.
   Bioinformatics 30(16):2389-2390, 2014.

.. [3] Greenfield, P., Droettboom, M., & Bray, E.,
   ASDF: A new data format for astronomy.
   Astronomy and Computing, 12: 240-251, 2015, doi:10.1016/j.ascom.2015.06.004

.. [4] Ouyang et al.,
   Deep learning massively accelerates super-resolution localization microscopy.
   Nat. Biotechnol. 2018, doi:10.1038/nbt.4106.

.. [5] https://github.com/imodpasteur/smlm-file-format

.. [6] Artur Speiser, Lucas-Raphael Müller, Philipp Hoess, Ulf Matti, Christopher J. Obara, Wesley R. Legant,
   Anna Kreshuk, Jakob H. Macke, Jonas Ries, and Srinivas C. Turaga,
   Deep learning enables fast and dense single-molecule localization with high accuracy.
   Nat Methods 18: 1082–1090, 2021, doi:10.1038/s41592-021-01236-x

.. [7] Ries, J.,
   SMAP: a modular super-resolution microscopy analysis platform for SMLM data.
   Nat Methods 17: 870–872, 2020, doi.org/10.1038/s41592-020-0938-1


Submodules:
-----------

.. autosummary::
   :toctree: ./

   io_locdata
   utilities
   rapidstorm_io
   thunderstorm_io
   elyra_io
   nanoimager_io
   asdf_io
   smlm_io
   decode_io
   smap_io

"""
from .io_locdata import *
from .utilities import *
from .rapidstorm_io import *
from .elyra_io import *
from .thunderstorm_io import *
from .nanoimager_io import *
from .asdf_io import *
from .smlm_io import *
from .decode_io import *
from .smap_io import *

__all__ = []
__all__.extend(io_locdata.__all__)
__all__.extend(utilities.__all__)
__all__.extend(rapidstorm_io.__all__)
__all__.extend(elyra_io.__all__)
__all__.extend(thunderstorm_io.__all__)
__all__.extend(nanoimager_io.__all__)
__all__.extend(asdf_io.__all__)
__all__.extend(smlm_io.__all__)
__all__.extend(decode_io.__all__)
__all__.extend(smap_io.__all__)



"""

Transform localization intensities.

Localization intensities can be given in counts, electrons or photons.
Here we provide often used transformation functions.

"""
import sys

import numpy as np
import pandas as pd

from locan.data.locdata import LocData
import locan.data.hulls
from locan.data.region import Region
from locan.simulation import simulate_uniform
from locan.data.metadata_utils import _modify_meta
from locan.dependencies import HAS_DEPENDENCY



__all__ = ['transform_counts_to_photons']


def _transform_counts_to_photons(intensities: np.ndarray, offset: float, gain: float, conversion_factor: float) \
        -> np.ndarray:
    """

    Parameters
    ----------
    intensities
    offset
    gain
    conversion_factor

    Returns
    -------

    """
    return intensities


import pytest

from locan.dependencies import HAS_DEPENDENCY
from locan import LocData
import locan.constants
import locan.locan_io.locdata.io_locdata as io
from locan.analysis import AccumulationClusterCheck
from locan.analysis.accumulation_analysis import _accumulation_cluster_check, \
    _accumulation_cluster_check_for_single_dataset

@pytest.fixture()
def locdata_blobs():
    return io.load_txt_file(path=locan.ROOT_DIR / 'tests/test_data/five_blobs.txt')

@pytest.fixture()
def locdata_blobs_3D():
    return io.load_txt_file(path=locan.ROOT_DIR / 'tests/test_data/five_blobs_3D.txt')


@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
def test___accumulation_cluster_check_for_single_dataset(locdata_blobs):
    res = _accumulation_cluster_check_for_single_dataset(locdata_blobs,
                                                         region_measure=locdata_blobs.bounding_box.region_measure,
                                                         hull='bb')
    assert len(res) == 3
    res = _accumulation_cluster_check_for_single_dataset(locdata_blobs,
                                                         region_measure=locdata_blobs.bounding_box.region_measure,
                                                         hull='ch')
    assert len(res) == 3


@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
def test__accumulation_cluster_check(locdata_blobs):
    results = _accumulation_cluster_check(locdata_blobs)
    assert len(results.columns) == 5

    results = _accumulation_cluster_check(locdata_blobs, n_loc=3, hull='bb', divide='sequential')
    assert len(results.columns) == 5

    results = _accumulation_cluster_check(locdata_blobs, n_loc=3, hull='ch', divide='sequential')
    assert len(results.columns) == 5

    results = _accumulation_cluster_check(locdata_blobs, region_measure=100, n_loc=3, divide='sequential')
    assert len(results.columns) == 5

    results = _accumulation_cluster_check(locdata_blobs, n_loc=[20, 30], divide='sequential')
    assert len(results.columns) == 5


@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
def test_AccumulationClusterCheck_empty(caplog):
    acc = AccumulationClusterCheck().compute(LocData())
    acc.plot()
    assert caplog.record_tuples == [('locan.analysis.accumulation_analysis', 30, 'Locdata is empty.')]


@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
def test_AccumulationClusterCheck(locdata_blobs):
    acc = AccumulationClusterCheck(algo_parameter=dict(min_cluster_size=3, allow_single_cluster=False),
                                    n_loc=5).compute(locdata_blobs)
    assert len(acc.results.columns) == 5
    acc = AccumulationClusterCheck(algo_parameter=dict(min_cluster_size=3, allow_single_cluster=False),
                                   n_loc=5, n_extrapolate=2).compute(locdata_blobs)
    assert len(acc.results.columns) == 5

    acc.plot()



from locan import ROOT_DIR, locdata_id


def test_root_directory():
    assert ROOT_DIR.is_dir()
    # print(ROOT_DIR.joinpath('tests/'))
    # print(type(ROOT_DIR))
    assert ROOT_DIR.joinpath('tests/').is_dir()

def test_locdata_id():
    assert locdata_id == 0



"""
Localization-density variation analysis to characterize localization cluster.

The existence of clusters is tested by analyzing variations in cluster area and localization density within clusters.
The analysis routine follows the ideas in [1]_ and [2]_.

Note
----
The analysis procedure is in an exploratory state and has not been fully developed and tested.

References
----------
.. [1] Baumgart F., Arnold AM., Leskovar K., Staszek K., Fölser M., Weghuber J., Stockinger H., Schütz GJ.,
   Varying label density allows artifact-free analysis of membrane-protein nanoclusters.
   Nat Methods. 2016 Aug;13(8):661-4. doi: 10.1038/nmeth.3897
.. [2] Spahn C., Herrmannsdörfer F., Kuner T., Heilemann M.
   Temporal accumulation analysis provides simplified artifact-free analysis of membrane-protein nanoclusters.
   Nat Methods. 2016 Nov 29;13(12):963-964. doi: 10.1038/nmeth.406

"""
import warnings
import logging

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from locan.data.locdata import LocData
from locan.analysis.analysis_base import _Analysis
from locan.data.cluster.clustering import cluster_hdbscan
from locan.data.filter import random_subset
from locan.data.hulls import ConvexHull


__all__ = ['AccumulationClusterCheck']

logger = logging.getLogger(__name__)

#### The algorithms


def _accumulation_cluster_check_for_single_dataset(locdata, region_measure, algorithm=cluster_hdbscan,
                                                   algo_parameter=None, hull='bb'):
    """
    Compute localization density, relative area coverage by the clusters (eta), average density of localizations
    within apparent clusters (rho) for a single localization dataset.
    """
    # localization density
    localization_density = len(locdata) / region_measure

    # compute clusters
    if algo_parameter is None:
        algo_parameter = {}
    noise, clust = algorithm(locdata, **algo_parameter)

    if len(clust)==0:
        # return localization_density, eta, rho
        return np.nan, np.nan, np.nan

    # compute cluster regions and densities
    if hull == 'bb':
        # Region_measure_bb has been computed upon instantiation
        if not 'region_measure_bb' in clust.data.columns:
            # return localization_density, eta, rho
            return np.nan, np.nan, np.nan

        else:
            # relative area coverage by the clusters
            eta = clust.data['region_measure_bb'].sum() / region_measure

            # average_localization_density_in_cluster
            rho = clust.data['localization_density_bb'].mean()

    elif hull == 'ch':
        # compute hulls
        Hs = [ConvexHull(ref.coordinates) for ref in clust.references]
        clust.dataframe = clust.dataframe.assign(region_measure_ch=[H.region_measure for H in Hs])

        localization_density_ch = clust.data['localization_count'] / clust.data['region_measure_ch']
        clust.dataframe = clust.dataframe.assign(localization_density_ch=localization_density_ch)

        # relative area coverage by the clusters
        eta = clust.data['region_measure_ch'].sum() / region_measure

        # average_localization_density_in_cluster
        rho = clust.data['localization_density_ch'].mean()

    else:
        raise TypeError('Computation for the specified hull is not implemented.')

    return localization_density, eta, rho


def _accumulation_cluster_check(locdata, region_measure='bb', algorithm=cluster_hdbscan,
                                algo_parameter=None, hull='bb', n_loc=10, divide='random', n_extrapolate=5):
    """
    Compute localization density, relative area coverage by the clusters (eta), average density of localizations
    within apparent clusters (rho) for the sequence of divided localization datasets.
    """
    # total region
    if isinstance(region_measure, str):
        region_measure_ = locdata.properties['region_measure_' + region_measure]
    else:
        region_measure_ = region_measure

    if isinstance(n_loc, (list, tuple, np.ndarray)):
        if max(n_loc) <= len(locdata):
            numbers_loc = n_loc
        else:
            raise ValueError('The bins must be smaller than the total number of localizations in locdata.')
    else:
        numbers_loc = np.linspace(0, len(locdata), n_loc+1, dtype=int)[1:]

    # take random subsets of localizations
    if divide == 'random':
        locdatas = [random_subset(locdata, n_points=n_pts) for n_pts in numbers_loc]
        for locd in locdatas:
            locd.reduce(reset_index=True)
    elif divide == 'sequential':
        selected_indices = [locdata.data.index[range(n_pts)].values for n_pts in numbers_loc]
        locdatas = [LocData.from_selection(locdata, indices=idx) for idx in selected_indices]
    else:
        raise TypeError(f'String input {divide} for divide is not valid.')

    results_ = [_accumulation_cluster_check_for_single_dataset(locd,
                                                               region_measure=region_measure_, algorithm=algorithm,
                                                               algo_parameter=algo_parameter, hull=hull)
                for locd in locdatas]

    # linear regression to extrapolate rho_0
    results_ = np.asarray(results_)
    idx = np.all(np.isfinite(results_), axis=1)
    x = results_[idx, 0]  # position 0 being localization_density
    y = results_[idx, 2]  # position 2 being rho

    fit_coefficients = np.polyfit(x[0:n_extrapolate], y[0:n_extrapolate], deg=1)
    rho_zero = fit_coefficients[-1]

    if rho_zero <= 0.:
        logger.warning('Extrapolation of rho yields a negative value.')
        rho_zero = 1

    # combine results
    results_ = [np.append(entry, [rho_zero, entry[2]/rho_zero]) for entry in results_]
    results = pd.DataFrame(data = results_,
                           columns=['localization_density', 'eta', 'rho', 'rho_0', 'rho/rho_0'])
    return results


##### The specific analysis classes

class AccumulationClusterCheck(_Analysis):
    """
    Check for the presence of clusters in localization data by analyzing variations in cluster area and localization
    density within clusters.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    region_measure : float, str
        Region measure (area or volume) for the support of locdata. String can be any of standard hull identifiere.
    algorithm : callable
        Clustering algorithm.
    algo_parameter : dict
        Dictionary with kwargs for `algorithm`.
    hull : str
        Hull computation that is used to compute cluster region measures (area or volume). The identifier string can
        be one of the defined hulls.
    n_loc : int, sequence of scalars
        If n_loc is an int, it defines the number of localization subsets into which the total number of localizations
        are distributed.
        If n_loc is a sequence, it defines the number of localizations used for each localization subset.
    divide: str
        Identifier to choose how to partition the localization data. For `random` localizations are selected randomly.
        For `sequential` localizations are selected as chuncks of increasing size always starting from the first
        element.
    n_extrapolate : int
        The number of rho values taken to extrapolate rho_zero.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        Data frame with  localization density, relative area coverage by the clusters (eta), average density of
        localizations within apparent clusters (rho), and rho normalized to the extrapolated value of rho for
        localization_density=0 (rho_zero). If the extrapolation of rho yields a negative value rho_zero is set to 1.
    """
    count = 0

    def __init__(self, meta=None, region_measure='bb', algorithm=cluster_hdbscan, algo_parameter=None,
                 hull='bb', n_loc=10, divide='random', n_extrapolate=5):
        super().__init__(meta=meta, region_measure=region_measure,
                         algorithm=algorithm, algo_parameter=algo_parameter, hull=hull, n_loc=n_loc, divide=divide,
                         n_extrapolate=n_extrapolate)

    def compute(self, locdata):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
          Localization data that might be clustered.

        Returns
        -------
        AccumulationClusterCheck
          Returns the `Analysis` object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        self.results = _accumulation_cluster_check(locdata, **self.parameter)
        return self

    def plot(self, ax=None, **kwargs):
        """
        Provide plot of results as :class:`matplotlib.axes.Axes` object.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        self.results.plot(x='eta', y='rho/rho_0', ax=ax, **kwargs)

        ax.set(title = '',
               xlabel = 'Relative clustered area',
               ylabel = 'Normalized localization density within cluster'
               )

        return ax



from pathlib import Path
import tempfile

import numpy as np
from pandas.testing import assert_frame_equal
from google.protobuf import json_format

import locan.constants
from locan.locan_io.locdata import manifest_pb2
from locan.locan_io.locdata.smlm_io import load_SMLM_manifest, load_SMLM_header, save_SMLM, \
    manifest_format_from_locdata, manifest_file_info_from_locdata, manifest_from_locdata, \
    load_SMLM_file, _change_upper_to_lower_keys


def test_manifest_format_from_locdata(locdata_2d):
    format = manifest_format_from_locdata(locdata_2d)
    assert isinstance(format, manifest_pb2.Format)
    #print(format)


def test_manifest_file_info_from_locdata(locdata_2d):
    file_info = manifest_file_info_from_locdata(locdata_2d)
    assert isinstance(file_info, manifest_pb2.FileInfo)
    #print(file_info)


def test_manifest_from_locdata(locdata_2d):
    manifest = manifest_from_locdata(locdata_2d)
    assert isinstance(manifest, manifest_pb2.Manifest)
    #print(manifest)

    # check for capital letters in manifest that should be introduced by manifest_pb2.Manifest
    json_string = json_format.MessageToJson(manifest,
                                            preserving_proto_field_name=True,
                                            including_default_value_fields=False)
    assert json_string
    assert "BINARY" in json_string
    assert "INT" in json_string

    # check that capital letters are eliminated by _change_upper_to_lower_keys in manifest_from_locdata
    manifest = manifest_from_locdata(locdata_2d, return_json_string=True)
    assert isinstance(manifest, str)
    assert "binary" in manifest
    assert "BINARY" not in manifest
    assert "int" in manifest
    assert "INT" not in manifest
    #print(manifest)


def test_get_correct_column_names_from_SMLM_header():
    columns = load_SMLM_header(path=locan.ROOT_DIR / 'tests/test_data/SMLM_dstorm_data.smlm')
    assert columns == ['original_index', 'position_x', 'local_background', 'chi_square',
                       'intensity', 'frame', 'position_y']


def test_load_SMLM_manifest():
    manifest = load_SMLM_manifest(path=locan.ROOT_DIR / 'tests/test_data/SMLM_dstorm_data.smlm')
    for key in ['format_version', 'formats', 'files']:
        assert key in manifest.keys()


def test_loading_SMLM_file():
    locdata = load_SMLM_file(
        path=locan.ROOT_DIR / 'tests/test_data/SMLM_dstorm_data.smlm',
        nrows=10)
    assert np.array_equal(locdata.data.columns,
                          ['original_index', 'position_x', 'local_background', 'chi_square',
                           'intensity', 'frame', 'position_y']
                          )
    assert len(locdata) == 10
    # print(locdata.data)


def test_save_and_load_smlm(locdata_2d):
    # introduce different dtypes in data
    locdata_2d.data["position_x"] = locdata_2d.data["position_x"].astype('float64')
    locdata_2d.data["position_y"] = locdata_2d.data["position_y"].astype('int32')
    locdata_2d.data["frame"] = locdata_2d.data["frame"].astype('int32')
    locdata_2d.data["intensity"] = locdata_2d.data["intensity"].astype('int32')
    assert all(locdata_2d.data.dtypes.values == ["float64", "int32", "int32", "int32"])
    # print(locdata_2d.data)
    # print(locdata_2d.data.dtypes)

    with tempfile.TemporaryDirectory() as tmp_directory:
        # save smlm file
        file_path = Path(tmp_directory) / 'locdata.smlm'
        save_SMLM(locdata_2d, path=file_path)

        # read back smlm manifest
        manifest = load_SMLM_manifest(path=file_path)
        for key in ['format_version', 'formats', 'files']:
            assert key in manifest.keys()

        # read back smlm file
        locdata = load_SMLM_file(path=file_path, convert=False)
        assert len(locdata) == len(locdata_2d)
        # assert (locdata.meta.identifier == locdata_2d.meta.identifier)
        assert (locdata.properties.keys() == locdata_2d.properties.keys())

        # print(locdata.data)
        # print(locdata.data.dtypes)
        # print(manifest_from_locdata(locdata_2d))
        assert_frame_equal(locdata_2d.data, locdata.data, check_dtype=True)

        # passing manifest
        manifest = manifest_from_locdata(locdata_2d, return_json_string=True)
        save_SMLM(locdata_2d, path=file_path, manifest=manifest)

        manifest = manifest_from_locdata(locdata_2d)
        save_SMLM(locdata_2d, path=file_path, manifest=manifest)

        # read back smlm manifest
        manifest = load_SMLM_manifest(path=file_path)
        for key in ['format_version', 'formats', 'files']:
            assert key in manifest.keys()

        # read back smlm file
        locdata = load_SMLM_file(path=file_path)
        assert_frame_equal(locdata_2d.data, locdata.data, check_dtype=False)



from locan import system_info, dependency_info, show_versions


def test__get_sys_info():
    sys_info = system_info()
    # print(sys_info)
    assert len(sys_info) == 12


def test___get_dependency_info():
    deps_info = dependency_info(other_dependencies=['locan'])
    # print(deps_info)
    assert 'boost-histogram' in deps_info
    assert 'locan' in deps_info


def test_show_versions_(capfd):
    show_versions(other_dependencies=['locan'])
    captured = capfd.readouterr()
    for header in ['\nLocan:', '\nPython:', '\nSystem:', '\nPython dependencies:',
                   'locan', 'node:', 'executable:']:
        assert header in captured.out

    show_versions(verbose=False, extra_dependencies=False)
    captured = capfd.readouterr()
    for header in ['\nLocan:', '\nPython:', '\nSystem:', '\nPython dependencies:']:
        assert header in captured.out
    for header in ['node:', 'executable:']:
        assert header not in captured.out



import pytest
import matplotlib.pyplot as plt  # needed for visual inspection

from locan import LocData
from locan.analysis import LocalizationProperty2d


def test_Localization_property_2d_empty(caplog):
    lprop = LocalizationProperty2d().compute(LocData())
    lprop.plot()
    lprop.plot_residuals()
    lprop.plot_deviation_from_mean()
    lprop.plot_deviation_from_median()
    lprop.report()
    assert caplog.record_tuples == [('locan.analysis.localization_property_2d', 30, 'Locdata is empty.'),
                                    ('locan.analysis.localization_property_2d', 30, 'No results available')]


def test_Localization_property_2d(capfd, locdata_rapidSTORM_2d):
    lprop = LocalizationProperty2d(meta=None, other_property='local_background', bin_size=1000)\
        .compute(locdata_rapidSTORM_2d)
    assert 'model_result' in lprop.results._fields
    assert lprop.results.model_result.params

    lprop.report()
    captured = capfd.readouterr()
    assert captured.out[:16] == "Fit results for:"

    lprop.plot()
    lprop.plot_residuals()
    lprop.plot_deviation_from_mean()
    lprop.plot_deviation_from_median()
    # plt.show()

    plt.close('all')



from pathlib import Path
import tempfile
from copy import deepcopy

import locan.constants
import numpy as np
import pytest
from pandas.testing import assert_frame_equal
from google.protobuf import json_format

import locan as lc
from locan.dependencies import HAS_DEPENDENCY
from locan import load_SMAP_header, load_SMAP_file, save_SMAP_csv, load_txt_file


pytestmark = pytest.mark.skipif(not HAS_DEPENDENCY["h5py"], reason="requires h5py.")


def test_load_SMAP_header(caplog):
    columns = load_SMAP_header(path=lc.ROOT_DIR / 'tests/test_data/smap_dstorm_data.mat')
    assert np.array_equal(columns,
                          ['LLrel', 'PSFxnm', 'PSFxpix', 'PSFynm', 'PSFypix', 'local_background', 'channel',
                           'filenumber', 'frame', 'iterations', 'locprecnm', 'locprecznm', 'logLikelihood',
                           'intensity', 'photerr', 'position_x', 'uncertainty_x', 'xpix', 'xpixerr', 'position_y',
                           'uncertainty_y', 'ypix', 'ypixerr', 'zerr', 'position_z']
                          )
    assert caplog.record_tuples[0] == ('locan.locan_io.locdata.utilities', 30,
                                       'Column LLrel is not a Locan property standard.')


# this is inactive because an empty .mat file is too large to be kept as test_data
# def test_loading_SMAP_file_empty_file():
#     locdata = load_SMAP_file(
#         path=lc.ROOT_DIR / 'tests/test_data/smap_dstorm_data_empty.mat',
#         nrows=10)
#     assert len(locdata) == 0


def test_loading_SMAP_file(caplog):
    locdata = load_SMAP_file(
        path=lc.ROOT_DIR / 'tests/test_data/smap_dstorm_data.mat',
        nrows=10)
    assert len(locdata) == 10
    assert np.array_equal(locdata.data.columns,
                          ['LLrel', 'PSFxnm', 'PSFxpix', 'PSFynm', 'PSFypix', 'local_background', 'channel',
                           'filenumber', 'frame', 'iterations', 'locprecnm', 'locprecznm', 'logLikelihood',
                           'intensity', 'photerr', 'position_x', 'uncertainty_x', 'xpix', 'xpixerr', 'position_y',
                           'uncertainty_y', 'ypix', 'ypixerr', 'zerr', 'position_z']
                          )
    assert caplog.record_tuples[0] == ('locan.locan_io.locdata.utilities', 30,
                                       'Column LLrel is not a Locan property standard.')


def test_save_and_load_smap_csv(locdata_2d):
    locdata = deepcopy(locdata_2d)
    # introduce different dtypes in data
    locdata.data["position_x"] = locdata.data["position_x"].astype('float64')
    locdata.data["position_y"] = locdata.data["position_y"].astype('int32')
    locdata.data["frame"] = locdata.data["frame"].astype('int32')
    locdata.data["intensity"] = locdata.data["intensity"].astype('int32')
    assert all(locdata.data.dtypes.values == ["float64", "int32", "int32", "int32"])
    # print(locdata.data)

    with tempfile.TemporaryDirectory() as tmp_directory:
        # save smap file
        file_path = Path(tmp_directory) / 'locdata.csv'
        save_SMAP_csv(locdata, path=file_path)

        # read back smap file
        locdata_ = load_txt_file(path=file_path, property_mapping=locan.constants.SMAP_KEYS)
        assert len(locdata_) == len(locdata)
        assert (locdata_.properties.keys() == locdata.properties.keys())

        # print(locdata_.data)
        # print(locdata_.data.dtypes)
        assert_frame_equal(locdata_.data, locdata.data, check_dtype=False)
        # todo: add convert kwarg to load_txt_file
        # assert_frame_equal(locdata_.data, locdata.data, check_dtype=True)



"""

Nearest-neighbor distance distribution analysis

Nearest-neighbor distance distributions provide information about deviations from a spatial homogeneous Poisson process
(i.e. complete spatial randomness, CSR).
Point-event distances are given by the distance between a random point (not being an event) and the nearest event.
The point-event distance distribution is estimated from a number of random sample points and plotted in comparison to
the analytical function for equal localization density.

For a homogeneous 2D Poisson process with intensity :math:`\\rho` (expected number of points per unit area) the distance
from a randomly chosen event to the nearest other event (nearest-neighbor distance) is distributed according to the
following probability density (pdf) or cumulative density function (cdf) [1]_:

.. math::

   pdf(w) &= 2 \\rho \\pi w \\ exp(- \\rho \\pi w^2)

   cdf(w) &= 1 - exp (- \\rho \\pi w^2)


The same distribution holds for point-event distances if events are distributed as a homogeneous Poisson process with
intensity :math:`\\rho`.

References
----------
.. [1] Philip M. Dixon, Nearest Neighbor Methods,
   Department of Statistics, Iowa State University,
   20 December 2001

"""
import logging

import numpy as np
import pandas as pd
from sklearn.neighbors import NearestNeighbors
from scipy import stats
import matplotlib.pyplot as plt

from locan.analysis.analysis_base import _Analysis, _list_parameters
from locan.configuration import N_JOBS


__all__ = ['NearestNeighborDistances']

logger = logging.getLogger(__name__)


#### The algorithms

def pdf_nnDistances_csr_2D(x, density):
    """
    Probability density function for nearest-neighbor distances of points distributed in 2D with complete spatial
    randomness.

    Parameters
    ----------
    x : float
        distance
    density : float
        density of points

    Returns
    -------
    float
        Probability density function pdf(x).
    """
    return 2 * density * np.pi * x * np.exp(-density * np.pi * x**2)


def pdf_nnDistances_csr_3D(x, density):
    """
    Probability density function for nearest-neighbor distances of points distributed in 3D with complete spatial
    randomness.

    Parameters
    ----------
    x : float
        distance
    density : float
        density of points

    Returns
    -------
    float
        Probability density function pdf(x).
    """
    a = (3 / 4 / np.pi / density)**(1/3)
    return 3 / a * (x/a)**2 * np.exp(-(x/a)**3)


def _nearest_neighbor_distances(points, k=1, other_points=None):

    if other_points is None:
        nn = NearestNeighbors(n_neighbors=k, metric='euclidean', n_jobs=N_JOBS).fit(points)
        distances, indices = nn.kneighbors()
    else:
        nn = NearestNeighbors(n_neighbors=k, metric='euclidean', n_jobs=N_JOBS).fit(other_points)
        distances, indices = nn.kneighbors(points)

    return pd.DataFrame({'nn_distance': distances[..., k-1], 'nn_index': indices[..., k-1]})


# The specific analysis classes

class NearestNeighborDistances(_Analysis):
    """
    Compute the k-nearest-neighbor distances within data or between data and other_data.

    The algorithm relies on sklearn.neighbors.NearestNeighbors.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    k : int
        Compute the kth nearest neighbor.


    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : numpy.ndarray, pandas.DataFrame
        Computed results.
    distribution_statistics : Distribution_stats, None
        Distribution parameters derived from MLE fitting of results.
    """
    count = 0

    def __init__(self, meta=None, k=1):
        super().__init__(meta=meta, k=k)
        self.dimension = None
        self.localization_density = None
        self.results = None
        self.distribution_statistics = None

    def compute(self, locdata, other_locdata=None):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
           Localization data.
        other_locdata : LocData
            Other localization data from which nearest neighbors are taken.

        Returns
        -------
        Analysis class
           Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        self.dimension = locdata.dimension
        # setting the localization density of locdata
        if other_locdata is None:
            self.localization_density = locdata.properties['localization_density_bb']
        else:
            if other_locdata.dimension != self.dimension:
                raise TypeError('Dimensions for locdata and other_locdata must be identical.')
            self.localization_density = other_locdata.properties['localization_density_bb']

        points = locdata.coordinates
        if other_locdata is None:
            other_points = None
        else:
            other_points = other_locdata.coordinates

        self.results = _nearest_neighbor_distances(points=points, **self.parameter, other_points=other_points)
        return self

    def fit_distributions(self, with_constraints=True):
        """
        Fit probability density functions to the distributions of `loc_property` values in the results
        using MLE (scipy.stats).

        If with_constraints is true we put the following constraints on the fit procedure:
        If distribution is expon then floc=np.min(self.analysis_class.results[self.loc_property].values).

        Parameters
        ----------
        distribution : str, scipy.stats.distribution
            Distribution model to fit.
        with_constraints : bool
            Flag to use predefined constraints on fit parameters.
        """
        if self:
            self.distribution_statistics = _DistributionFits(self)
            self.distribution_statistics.fit(with_constraints=with_constraints)
        else:
            logger.warning('No results available to fit.')

    def hist(self, ax=None, bins='auto', density=True, fit=False, **kwargs):
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing hist(results).

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image.
        bins : int, list, 'auto'
            Bin specification as used in :func:`matplotlib.hist`
        density : bool
            Flag for normalization as used in matplotlib.hist. True returns probability density function; None returns
            counts.
        fit : bool
            Flag indicating to fit pdf of nearest-neighbor distances under complete spatial randomness.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        values, bin_values, patches = ax.hist(self.results['nn_distance'], bins=bins, density=density, label='data')
        x_data = (bin_values[:-1] + bin_values[1:]) / 2

        if self.dimension == 2:
            ax.plot(x_data, pdf_nnDistances_csr_2D(x_data, self.localization_density), 'r-', label='CSR', **kwargs)
        elif self.dimension == 3:
            ax.plot(x_data, pdf_nnDistances_csr_3D(x_data, self.localization_density), 'r-', label='CSR', **kwargs)
        else:
            logger.warning(f'No analytic probability density function for {self.dimension} dimensions available.')

        # fit distributions:
        if fit:
            if isinstance(self.distribution_statistics, _DistributionFits):
                self.distribution_statistics.plot(ax=ax)
            else:
                self.fit_distributions()
                self.distribution_statistics.plot(ax=ax)

        ax.set(title='k-Nearest Neigbor Distances\n'+' (k = ' + str(self.parameter['k'])+')',
               xlabel='distance (nm)',
               ylabel='pdf' if density else 'counts'
               )
        ax.legend(loc='best')

        return ax


#### Auxiliary functions and classes

class NNDistances_csr_2d(stats.rv_continuous):
    """
    Continuous distribution function for nearest-neighbor distances of points distributed in 2D
    under complete spatial randomness.

    Parameters
    ----------
    density : float
        Shape parameter `density`, being the density of points.
    """

    def _pdf(self, x, density):
        return 2 * density * np.pi * x * np.exp(-density * np.pi * x**2)


class NNDistances_csr_3d(stats.rv_continuous):
    """
    Continuous distribution function for nearest-neighbor distances of points distributed in 3D
    under complete spatial randomness.

    Parameters
    ----------
    density : float
        Shape parameter `density`, being the density of points.
    """

    def _pdf(self, x, density):
        a = (3 / 4 / np.pi / density) ** (1 / 3)
        return 3 / a * (x / a) ** 2 * np.exp(-(x / a) ** 3)


class _DistributionFits:
    """
    Handle for distribution fits.

    This class is typically instantiated by LocalizationProperty methods.
    It holds the statistical parameters derived by fitting the result distributions using MLE (scipy.stats).
    Statistical parameters are defined as described in
    :ref:(https://docs.scipy.org/doc/scipy/reference/tutorial/stats/continuous.html)

    Parameters
    ----------
    analyis_class : LocalizationPrecision
        The analysis class with result data to fit.

    Attributes
    ----------
    analyis_class : LocalizationPrecision
        The analysis class with result data to fit.
    loc_property : str
        The LocData property for which to fit an appropriate distribution
    distribution : str, scipy.stats.distribution
        Distribution model to fit.
    parameters : list of str
        Free parameters in `distribution`.
    """
    def __init__(self, analysis_class):
        self.analysis_class = analysis_class
        self.loc_property = 'nn_distance'
        self.distribution = None
        self.parameters = []

    def fit(self, with_constraints=True, **kwargs):
        """
        Fit model function to analysis_class.results.

        If with_constraints is true (default) we put the following constraints on the fit procedure:
        loc=0, scale=1

        Parameters
        ----------
        distribution : str, scipy.stats.distribution
            Distribution model to fit.
        with_constraints : bool
            Flag to use predefined constraints on fit parameters.
        kwargs : dict
            Other parameters passed to the `distribution.fit()` method.
        """
        if self.analysis_class.dimension == 2:
            self.distribution = NNDistances_csr_2d(name='NNDistances_csr_2d', a=0.)
        elif self.analysis_class.dimension == 3:
            self.distribution = NNDistances_csr_3d(name='NNDistances_csr_3d', a=0.)
        else:
            logger.warning(f'No fit model for {self.analysis_class.dimension} dimensions available.')
            return

        self.parameters = [name.strip() for name in self.distribution.shapes.split(',')]
        self.parameters += ['loc', 'scale']

        if with_constraints:
            kwargs_ = dict(dict(floc=0, fscale=1), **kwargs)
        else:
            kwargs_ = kwargs

        fit_results = self.distribution.fit(data=self.analysis_class.results[self.loc_property].values, **kwargs_)
        for parameter, result in zip(self.parameters, fit_results):
            setattr(self, parameter, result)

    def plot(self, ax=None, **kwargs):
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the probability distribution functions of fitted
        results.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.distribution is None:
            return ax

        # plot fit curve
        x_values = np.linspace(self.distribution.ppf(0.001, **self.parameter_dict()),
                               self.distribution.ppf(0.999, **self.parameter_dict()), 100)
        ax.plot(x_values, self.distribution.pdf(x_values, **self.parameter_dict()), 'r-',
                **dict(dict(lw=3, alpha=0.6, label=str(self.distribution.name) + ' pdf'), **kwargs))
        return ax

    def parameter_dict(self):
        """ Dictionary of fitted parameters. """
        return {k: self.__dict__[k] for k in self.parameters}



"""

Methods for clustering localization data in LocData objects.

"""

import sys
from copy import copy, deepcopy

import numpy as np
import pandas as pd
from sklearn.neighbors import NearestNeighbors
from scipy import stats

from sklearn.cluster import DBSCAN

from locan import HAS_DEPENDENCY, needs_package
if HAS_DEPENDENCY["hdbscan"]: from hdbscan import HDBSCAN

from locan.data.locdata import LocData
from locan.configuration import N_JOBS


__all__ = ['cluster_hdbscan', 'cluster_dbscan']


@needs_package("hdbscan")
def cluster_hdbscan(locdata, min_cluster_size=5, loc_properties=None, allow_single_cluster=False, **kwargs):
    """
    Cluster localizations in locdata using the hdbscan clustering algorithm.

    Parameters
    ----------
    locdata : LocData
        specifying the localization data on which to perform the manipulation.
    loc_properties : list of string, None
        The LocData properties to be used for clustering. If None, locdata.coordinates will be used.
    min_cluster_size : int
        minimumm cluster size in HDBSCAN algorithm (default: 5)
    allow_single_cluster : bool
        allowing to return single cluster (default: False)
    kwargs : dict
        Other parameters passed to `hdbscan.HDBSCAN`.

    Returns
    -------
    tuple (LocData, LocData)
        A tuple with noise and cluster.
        The first LocData object is a selection of all localizations that are defined as noise,
        in other words all localizations that are not part of any cluster.
        The second LocData object is a LocData instance assembling all generated selections (i.e. localization cluster).
    """
    parameter = locals()

    if len(locdata) == 0:
        locdata_noise = LocData()
        collection = LocData()

    if len(locdata) < min_cluster_size:
        locdata_noise = copy(locdata)
        collection = LocData()

    else:
        if loc_properties is None:
            fit_data = locdata.coordinates
        else:
            fit_data = locdata.data[loc_properties]

        labels = HDBSCAN(
            min_cluster_size=min_cluster_size,
            allow_single_cluster=allow_single_cluster,
            gen_min_span_tree=False,
            **kwargs
        ).fit_predict(fit_data)

        grouped = locdata.data.groupby(labels)
        locdata_index_labels = [locdata.data.index[idxs] for idxs in grouped.indices.values()]
        selections = [LocData.from_selection(locdata=locdata, indices=idxs) for idxs in locdata_index_labels]

        try:
            grouped.get_group(-1)
            locdata_noise = selections[0]
            collection = LocData.from_collection(selections[1:])
        except KeyError:
            locdata_noise = None
            collection = LocData.from_collection(selections)

    # set regions
    if locdata_noise:
        locdata_noise.region = locdata.region
    if collection:
        collection.region = locdata.region

    # metadata
    if locdata_noise:
        del locdata_noise.meta.history[:]
        locdata_noise.meta.history.add(name=sys._getframe().f_code.co_name, parameter=str(parameter))
    del collection.meta.history[:]
    collection.meta.history.add(name=sys._getframe().f_code.co_name, parameter=str(parameter))

    return locdata_noise, collection


def cluster_dbscan(locdata, eps=20, min_samples=5, loc_properties=None, **kwargs):
    """
    Cluster localizations in locdata using the dbscan clustering algorithm as implemented in sklearn.

    Parameters
    ----------
    locdata : LocData
        specifying the localization data on which to perform the manipulation.
    eps : float
        The maximum distance between two samples for them to be considered as in the same neighborhood.
    min_samples : int
        The number of samples in a neighborhood for a point to be considered as a core point.
        This includes the point itself.
    loc_properties : list of string, None
        The LocData properties to be used for clustering. If None, locdata.coordinates will be used.
    kwargs : dict
        Other parameters passed to `sklearn.cluster.DBSCAN`.

    Returns
    -------
    tuple (LocData, LocData)
        A tuple with noise and cluster.
        The first LocData object is a selection of all localizations that are defined as noise,
        in other words all localizations that are not part of any cluster.
        The second LocData object is a LocData instance assembling all generated selections (i.e. localization cluster).
    """
    parameter = locals()

    if len(locdata) == 0:
        locdata_noise = LocData()
        collection = LocData()

    else:
        if loc_properties is None:
            fit_data = locdata.coordinates
        else:
            fit_data = locdata.data[loc_properties]

        labels = DBSCAN(
            eps=eps, min_samples=min_samples, n_jobs=N_JOBS, **kwargs
        ).fit_predict(fit_data)

        grouped = locdata.data.groupby(labels)
        locdata_index_labels = [locdata.data.index[idxs] for idxs in grouped.indices.values()]
        selections = [LocData.from_selection(locdata=locdata, indices=idxs) for idxs in locdata_index_labels]

        try:
            grouped.get_group(-1)
            locdata_noise = selections[0]
            collection = LocData.from_collection(selections[1:])
        except KeyError:
            locdata_noise = None
            collection = LocData.from_collection(selections)

    # set regions
    if locdata_noise:
        locdata_noise.region = locdata.region
    if collection:
        collection.region = locdata.region

    # metadata
    if locdata_noise:
        del locdata_noise.meta.history[:]
        locdata_noise.meta.history.add(name=sys._getframe().f_code.co_name, parameter=str(parameter))
    del collection.meta.history[:]
    collection.meta.history.add(name=sys._getframe().f_code.co_name, parameter=str(parameter))

    return locdata_noise, collection



import pytest
import matplotlib.pyplot as plt  # needed for visual inspection

from locan import LocData
from locan.analysis import LocalizationPropertyCorrelations


def test_Localization_property_correlations_empty(caplog):
    lpcorr = LocalizationPropertyCorrelations().compute(LocData())
    lpcorr.plot()
    assert caplog.record_tuples == [('locan.analysis.localization_property_correlations', 30, 'Locdata is empty.')]


def test_Localization_property_correlations(locdata_rapidSTORM_2d):
    lpcorr = LocalizationPropertyCorrelations().compute(locdata=locdata_rapidSTORM_2d)
    # print(lpcorr.results)
    for i in range(len(locdata_rapidSTORM_2d.data.columns)):
        assert lpcorr.results.iloc[i, i] == pytest.approx(1)

    lpcorr = LocalizationPropertyCorrelations(loc_properties=['intensity', 'local_background'])\
        .compute(locdata=locdata_rapidSTORM_2d)
    for i in range(2):
        assert lpcorr.results.iloc[i, i] == pytest.approx(1)

    lpcorr.plot()
    # lpcorr.plot(cbar=False, vmin=0)
    # plt.show()

    plt.close('all')



import os
from pathlib import Path

import pytest

from locan.dependencies import HAS_DEPENDENCY
from locan import ROOT_DIR
from locan.scripts.script_napari import sc_napari

if HAS_DEPENDENCY["napari"]: import napari


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_napari():
    viewer = napari.Viewer(show=False)
    assert viewer


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_script_napari():
    path = Path(ROOT_DIR / 'tests/test_data/five_blobs.txt')
    assert path.exists()
    sc_napari(file_path=str(path), file_type=1, bin_size=20, rescale=None)


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_script_napari_from_sys():
    path = Path(ROOT_DIR / 'tests/test_data/five_blobs.txt')
    assert path.exists()
    exit_status = os.system(f'locan napari -f {str(path)} -t 1 --bin_size 20 --rescale None')
    assert exit_status == 0



"""

Test the locan package.

This module includes unit tests for all modules within the locan package.
The tests are organized following the subpackage structure of locan.

"""
from locan import ROOT_DIR

__all__ = ['test']


def test(args=None):
    """
    Running tests with pytest.

    Parameters
    ----------
    args : str, list of str
        Parameters passed to :func:`pytest.main`
    """
    try:
        import pytest
    except ImportError:
        raise ImportError("Need pytest to run tests.")

    extra_args = ["-m not gui and not visual and not requires_datasets"]

    if args is None:
        pass  # extra_args = []
    elif isinstance(args, list):
        extra_args.extend(args)  # extra_args = args
    else:
        extra_args.append(args)  # extra_args = [args]

    test_directory = ROOT_DIR.joinpath('tests').as_posix()
    extra_args.append(test_directory)
    print(f'running: pytest {" ".join(extra_args)}')
    return pytest.main(extra_args)



"""

Track localizations.

This module provides functions for tracking localizations (i.e. clustering localization data in time).
The functions take LocData as input and compute new LocData objects.
It makes use of the trackpy package.

"""

import sys

import numpy as np
import pandas as pd

from locan.data.locdata import LocData
from locan.dependencies import HAS_DEPENDENCY, needs_package

if HAS_DEPENDENCY["trackpy"]: from trackpy import link_df


__all__ = ['link_locdata', 'track']


@needs_package("trackpy")
def link_locdata(locdata, search_range=40, memory=0, **kwargs):
    """
    Track localizations, i.e. cluster localizations in time when nearby in successive frames.
    This function applies the trackpy linking method to LocData objects.

    Parameters
    ----------
    locdata : LocData
        Localization data on which to perform the manipulation.
    search_range : float, tuple
        The maximum distance features can move between frames,
        optionally per dimension
    memory : int
        The maximum number of frames during which a feature can vanish,
        then reappear nearby, and be considered the same particle. 0 by default.
    kwargs :
        Other parameters passed to trackpy.link_df().

    Returns
    -------
    pandas Series
        A series named 'Track' referring to the track number.

    Note
    ----
    In order to switch off the printout from :func:`trackpy.link` and increase performance use
    :func:`trackpy.quiet()` to silence the logging outputs.
    """
    df = link_df(locdata.data, search_range=search_range, memory=memory, pos_columns=locdata.coordinate_labels,
                    t_column='frame', **kwargs)
    return_series = df['particle']
    return_series.name = 'track'
    return return_series


def track(locdata, search_range=40, memory=0, **kwargs):
    """
    Cluster (in time) localizations in LocData that are nearby in successive frames. Clustered localizations are
    identified by the trackpy linking method.

    The new locdata object carries properties with the same name as the original `locata`.
    They are computed as sum for `intensity`, as the first value for `frame`, and as mean for all other properties.

    Parameters
    ----------
    locdata : LocData
        Localization data on which to perform the manipulation.
    search_range : float, tuple
        The maximum distance features can move between frames, optionally per dimension
    memory : int
        The maximum number of frames during which a feature can vanish, then reappear nearby,
        and be considered the same particle.
    kwargs :
        Other parameters passed to trackpy.link_df.

    Returns
    -------
    Locdata, pandas Series
        A new LocData instance assembling all generated selections (i.e. localization cluster).
        A series named 'Track' referring to the track number.

    Note
    ----
    In order to switch off the printout from :func:`trackpy.link` and increase performance use
    :func:`trackpy.quiet()` to silence the logging outputs.
    """
    parameter = locals()

    track_series = link_locdata(locdata, search_range, memory, **kwargs)
    grouped = track_series.groupby(track_series)
    selections = [LocData.from_selection(locdata=locdata, indices=group.index.values) for _, group in grouped]

    for selection in selections:
        for column in locdata.data.columns:
            if 'position' not in column and 'intensity' != column and 'frame' != column:
                column_mean = getattr(selection.data, column).mean()
                selection.properties.update({column: column_mean})
            if column == 'intensity':
                intensity_mean = selection.data.intensity.mean()
                selection.properties.update({'intensity_mean': intensity_mean})
                intensity_sum = selection.data.intensity.sum()
                selection.properties.update({'intensity': intensity_sum})
            if column == 'frame':
                frame = selection.data.frame.iloc[0]
                selection.properties.update({'frame': frame})

    collection = LocData.from_collection(selections)

    # metadata
    del locdata.meta.history[:]
    locdata.meta.history.add(name=sys._getframe().f_code.co_name, parameter=str(parameter))

    return collection, track_series
